# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Tests for local.py, an OCW (netCDF, HDF5) file loading library."""

# Needed Python 2/3 urllib compatibility
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

import datetime
import unittest
import os
import netCDF4
import numpy as np

import ocw.data_source.local as local


class TestLoadFile(unittest.TestCase):
    """Tests for load_file method."""

    @classmethod
    def setUpClass(cls):
        """Prepare a netCDF file once to use for all tests."""
        cls.file_path = create_netcdf_file()

    @classmethod
    def tearDownClass(cls):
        """Remove the no longer needed testing file at the end of the tests."""
        os.remove(cls.file_path)

    def setUp(self):
        """Open and read in attributes of netCDF test file."""
        self.netcdf_file = netCDF4.Dataset(self.file_path, 'r')
        self.latitudes = self.netcdf_file.variables['latitude'][:]
        self.longitudes = self.netcdf_file.variables['longitude'][:]
        self.alt_lats = self.netcdf_file.variables['alt_lat'][:]
        self.alt_lons = self.netcdf_file.variables['alt_lon'][:]
        self.values = self.netcdf_file.variables['value'][:]
        self.variable_name_list = ['latitude', 'longitude', 'time', 'level',
                                   'value']
        self.possible_value_name = ['latitude', 'longitude', 'time', 'level']

    def tearDown(self):
        """Close file object so that it may be re-read in the next test."""
        self.netcdf_file.close()

    def test_load_invalid_file_path(self):
        """To test load_file an invalid path raises an exception."""
        self.invalid_netcdf_path = '/invalid/path'
        with self.assertRaises(ValueError):
            local.load_file(file_path=self.invalid_netcdf_path,
                            variable_name='test variable')

    def test_function_load_file_lats(self):
        """Test load_file function for latitudes."""
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value").lats, self.latitudes)

    def test_function_load_file_lons(self):
        """Test load_file function for longitudes."""
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value").lons, self.longitudes)

    def test_function_load_file_times(self):
        """Test load_file function for times."""
        new_times = datetime.datetime(2001, 1, 1), datetime.datetime(
            2001, 2, 1), datetime.datetime(2001, 3, 1)
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value").times, new_times)

    def test_function_load_file_alt_lats(self):
        """Test load_file function for lats with different variable names."""
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value", lat_name="alt_lat").lats, self.alt_lats)

    def test_function_load_file_alt_lons(self):
        """Test load_file function for lons with different variable names."""
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value", lon_name="alt_lon").lons, self.alt_lons)

    def test_function_load_file_alt_times(self):
        """Test load_file function for times with different variable names."""
        new_times = datetime.datetime(2001, 4, 1), datetime.datetime(
            2001, 5, 1), datetime.datetime(2001, 6, 1)
        np.testing.assert_array_equal(local.load_file(
            self.file_path, "value", time_name="alt_time").times, new_times)

    def test_function_load_file_values(self):
        """Test load_file function for values."""
        new_values = self.values[:, 0, :, :]
        self.assertTrue(np.allclose(local.load_file(
            self.file_path, "value").values, new_values))

    def test_custom_dataset_name(self):
        """Test adding a custom name to a dataset."""
        dataset = local.load_file(self.file_path, 'value', name='foo')
        self.assertEqual(dataset.name, 'foo')

    def test_dataset_origin(self):
        """Test that dataset origin is local."""
        dataset = local.load_file(self.file_path, 'value', elevation_index=1)
        expected_keys = {'source', 'path', 'lat_name', 'lon_name', 'time_name',
                         'elevation_index'}
        self.assertEqual(set(dataset.origin.keys()), expected_keys)
        self.assertEqual(dataset.origin['source'], 'local')


class TestLoadMultipleFiles(unittest.TestCase):
    """Tests for the load_multiple_files method."""

    @classmethod
    def setUpClass(cls):
        """Prepare a netCDF file once to use for all tests."""
        cls.file_path = create_netcdf_file()

    @classmethod
    def tearDownClass(cls):
        """Remove the no longer needed testing file at the end of the tests."""
        os.remove(cls.file_path)

    def setUp(self):
        """Open and read in attributes of netCDF test file."""
        self.netcdf_file = netCDF4.Dataset(self.file_path, 'r')
        self.latitudes = self.netcdf_file.variables['latitude'][:]
        self.longitudes = self.netcdf_file.variables['longitude'][:]
        self.values = self.netcdf_file.variables['value'][:]
        self.variable_name_list = ['latitude',
                                   'longitude', 'time', 'level', 'value']
        self.possible_value_name = ['latitude', 'longitude', 'time', 'level']

    def tearDown(self):
        """Close file object so that it may be re-read in the next test."""
        self.netcdf_file.close()

    def test_function_load_multiple_files_data_name(self):
        """Test load_multiple_file function for dataset name."""
        dataset = local.load_multiple_files(self.file_path, "value")
        self.assertEqual([dataset[0].name], [''])

    def test_function_load_multiple_files_lons(self):
        """Test load_multiple_file function for longitudes."""
        dataset = local.load_multiple_files(self.file_path, "value")
        np.testing.assert_array_equal(dataset[0].lons, self.longitudes)

    def test_function_load_multiple_files_times(self):
        """Test load_multiple_files function for times."""
        dataset = local.load_multiple_files(self.file_path, "value")

        new_times = datetime.datetime(2001, 1, 1), datetime.datetime(
            2001, 2, 1), datetime.datetime(2001, 3, 1)
        np.testing.assert_array_equal(dataset[0].times, new_times)

    def test_function_load_multiple_files_values(self):
        """Test load_multiple_files function for values."""
        new_values = self.values[:, 0, :, :]
        dataset = local.load_multiple_files(
            self.file_path, "value")
        self.assertTrue(np.allclose(dataset[0].values, new_values))

    def test_load_multiple_files_custom_dataset_name(self):
        """Test adding a custom name to a dataset."""
        dataset = local.load_multiple_files(self.file_path,
                                            "value",
                                            generic_dataset_name=True,
                                            dataset_name=['foo'])
        self.assertEqual(dataset[0].name, 'foo')

    def test_dataset_origin(self):
        """Test that dataset origin is local."""
        dataset = local.load_multiple_files(self.file_path, 'value')
        expected_keys = {'source', 'path', 'lat_name', 'lon_name', 'time_name'}
        self.assertEqual(set(dataset[0].origin.keys()), expected_keys)
        self.assertEqual(dataset[0].origin['source'], 'local')


class TestLoadDatasetFromMultipleNetcdfFiles(unittest.TestCase):
    """Tests for load_dataset_from_multiple_netcdf_files method."""

    @classmethod
    def setUpClass(cls):
        """Create, read in, and record attributes of a netCDF file for tests."""
        cls.file_path = create_netcdf_file()
        cls.netcdf_file = netCDF4.Dataset(cls.file_path, 'r')
        cls.latitudes = cls.netcdf_file.variables['latitude'][:]
        cls.longitudes = cls.netcdf_file.variables['longitude'][:]
        cls.alt_lats = cls.netcdf_file.variables['alt_lat'][:]
        cls.alt_lons = cls.netcdf_file.variables['alt_lon'][:]
        cls.values = cls.netcdf_file.variables['value'][:]
        cls.variable_name_list = ['latitude', 'longitude', 'time', 'level',
                                  'value']
        cls.possible_value_name = ['latitude', 'longitude', 'time', 'level']
        cls.dataset = local.load_dataset_from_multiple_netcdf_files(
            variable_name='value',
            file_path='',
            filename_pattern=[cls.file_path])
        cls.alt_dataset = local.load_dataset_from_multiple_netcdf_files(
            variable_name='value',
            lat_name='alt_lat',
            lon_name='alt_lon',
            time_name='alt_time',
            file_path='',
            filename_pattern=[cls.file_path])

    @classmethod
    def tearDownClass(cls):
        """Remove the no longer needed testing file at the end of the tests."""
        cls.netcdf_file.close()
        os.remove(cls.file_path)

    def test_variable_name(self):
        """Test that dataset contains a variable value."""
        self.assertEqual(self.dataset.variable, 'value')

    def test_function_load_dataset_from_multiple_netcdf_files_lats(self):
        """Test load_multiple_files function for times."""
        np.testing.assert_array_equal(self.dataset.lats, self.latitudes)

    def test_function_load_dataset_from_multiple_netcdf_files_lons(self):
        """Test load_multiple_files function for times."""
        np.testing.assert_array_equal(self.dataset.lons, self.longitudes)

    def test_function_load_dataset_from_multiple_netcdf_files_times(self):
        """Test load_multiple_files function for times."""
        new_times = datetime.datetime(2001, 1, 1), datetime.datetime(
            2001, 2, 1), datetime.datetime(2001, 3, 1)
        np.testing.assert_array_equal(self.dataset.times, new_times)

    def test_function_load_dataset_from_multiple_netcdf_files_alt_lats(self):
        """Test load_multiple_files function for non-default lats."""
        np.testing.assert_array_equal(self.alt_dataset.lats, self.alt_lats)

    def test_function_load_dataset_from_multiple_netcdf_files_alt_lons(self):
        """Test load_multiple_files function for non-default lons."""
        np.testing.assert_array_equal(self.alt_dataset.lons, self.alt_lons)

    def test_function_load_dataset_from_multiple_netcdf_files_alt_times(self):
        """Test load_multiple_files function for non-default times."""
        new_times = datetime.datetime(2001, 4, 1), datetime.datetime(
            2001, 5, 1), datetime.datetime(2001, 6, 1)
        np.testing.assert_array_equal(self.alt_dataset.times, new_times)

    def test_function_load_dataset_from_multiple_netcdf_files_values(self):
        """Test load_multiple_files function for values."""
        new_values = self.values[:, 0, :, :]
        self.assertTrue(np.allclose(self.dataset.values, new_values))


class TestGetNetcdfVariableNames(unittest.TestCase):
    """Tests for _get_netcdf_variable_name method retrieving variables.

    TestGetNetcdfVariableNames.nc" is a subset of data from
        https://zipper.jpl.nasa.gov/dist/AFRICA_KNMI-RACMO2.2b_CTL_ERAINT_MM_50km_1989-2008_tasmax.nc
        Test data obtained with:
        ncea -d time,0,0 AFRICA_KNMI-[...]_tasmax.nc \
         TestGetNetcdfVariableNames.nc
    """

    @classmethod
    def setUpClass(cls):
        """Create a netCDF file with invalid dimensions for tests."""
        cls.test_model = "TestGetNetcdfVariableNames.nc"
        cls.invalid_netcdf_path = create_invalid_dimensions_netcdf_file()

    def setUp(self):
        """Open a valid netCDF file for use in the test."""
        self.netcdf = netCDF4.Dataset(self.test_model, mode='r')

    def tearDown(self):
        """Close file object so that it may be re-read in the next test."""
        self.netcdf.close()

    def test_valid_latitude(self):
        """Test that a latitude variable (rlat) can be found in netCDF file."""
        self.lat = local._get_netcdf_variable_name(local.LAT_NAMES,
                                                   self.netcdf,
                                                   "tasmax")
        self.assertEquals(self.lat, "rlat")

    def test_invalid_dimension_latitude(self):
        """Test than an invalid  latitude variable can be found in file."""
        self.netcdf = netCDF4.Dataset(self.invalid_netcdf_path, mode='r')
        self.lat = local._get_netcdf_variable_name(local.LAT_NAMES,
                                                   self.netcdf,
                                                   "value")
        self.assertEquals(self.lat, "latitude")

    def test_dimension_variable_name_mismatch(self):
        """Test that mismatched latitude variables are found as latitude."""
        self.netcdf = netCDF4.Dataset(self.invalid_netcdf_path, mode='r')
        self.lat = local._get_netcdf_variable_name(
            ["lat_dim"] + local.LAT_NAMES,
            self.netcdf,
            "value")
        self.assertEquals(self.lat, "latitude")

    def test_no_match_latitude(self):
        """Test that retrieving a nonexistent variable name raises exception."""
        with self.assertRaises(ValueError):
            self.lat = local._get_netcdf_variable_name(['notAVarName'],
                                                       self.netcdf,
                                                       "tasmax")


def create_netcdf_file():
    """Create a temporary netCDF file with data used for testing."""
    # To create the temporary netCDF file
    file_path = '/tmp/temporaryNetcdf.nc'
    netcdf_file = netCDF4.Dataset(file_path, 'w', format='NETCDF4')
    # To create dimensions
    netcdf_file.createDimension('lat_dim', 5)
    netcdf_file.createDimension('lon_dim', 5)
    netcdf_file.createDimension('time_dim', 3)
    netcdf_file.createDimension('level_dim', 2)
    # To create variables
    latitudes = netcdf_file.createVariable('latitude', 'd', ('lat_dim',))
    longitudes = netcdf_file.createVariable('longitude', 'd', ('lon_dim',))
    times = netcdf_file.createVariable('time', 'd', ('time_dim',))
    # unusual variable names to test optional arguments for Dataset constructor
    alt_lats = netcdf_file.createVariable('alt_lat', 'd', ('lat_dim',))
    alt_lons = netcdf_file.createVariable('alt_lon', 'd', ('lon_dim',))
    alt_times = netcdf_file.createVariable('alt_time', 'd', ('time_dim',))
    levels = netcdf_file.createVariable('level', 'd', ('level_dim',))
    values = netcdf_file.createVariable('value', 'd',
                                        ('time_dim',
                                         'level_dim',
                                         'lat_dim',
                                         'lon_dim')
                                        )

    # To latitudes and longitudes for five values
    latitudes_data = np.arange(5.)
    longitudes_data = np.arange(150., 155.)
    # Three months of data.
    times_data = np.arange(3)
    # Two levels
    levels_data = [100, 200]
    # Create 150 values
    values_data = np.array([i for i in range(150)])
    # Reshape values to 4D array (level, time, lats, lons)
    values_data = values_data.reshape(len(times_data), len(
        levels_data), len(latitudes_data), len(longitudes_data))

    # Ingest values to netCDF file
    latitudes[:] = latitudes_data
    longitudes[:] = longitudes_data
    times[:] = times_data
    alt_lats[:] = latitudes_data + 10
    alt_lons[:] = longitudes_data - 10
    alt_times[:] = times_data
    levels[:] = levels_data
    values[:] = values_data
    # Assign time info to time variable
    netcdf_file.variables['time'].units = 'months since 2001-01-01 00:00:00'
    netcdf_file.variables[
        'alt_time'].units = 'months since 2001-04-01 00:00:00'
    netcdf_file.variables['value'].units = 'foo_units'
    netcdf_file.close()
    return file_path


def create_invalid_dimensions_netcdf_file():
    """Create a temporary netCDF file with invalid dimensions for testing."""
    # To create the temporary netCDF file
    file_path = '/tmp/temporaryNetcdf.nc'
    netcdf_file = netCDF4.Dataset(file_path, 'w', format='NETCDF4')
    # To create dimensions
    netcdf_file.createDimension('lat_dim', 5)
    netcdf_file.createDimension('lon_dim', 5)
    netcdf_file.createDimension('time_dim', 3)
    netcdf_file.createDimension('level_dim', 2)
    # To create variables
    latitudes = netcdf_file.createVariable('latitude', 'd', ('lat_dim',))
    longitudes = netcdf_file.createVariable('longitude', 'd', ('lon_dim',))
    times = netcdf_file.createVariable('time', 'd', ('time_dim',))
    levels = netcdf_file.createVariable('level', 'd', ('level_dim',))
    values = netcdf_file.createVariable('value',
                                        'd',
                                        ('level_dim',
                                         'time_dim',
                                         'lat_dim',
                                         'lon_dim')
                                        )
    # To latitudes and longitudes for five values
    flatitudes = list(range(0, 5))
    flongitudes = list(range(200, 205))
    # Three months of data
    ftimes = list(range(3))
    # Two levels
    flevels = [100, 200]
    # Create 150 values
    fvalues = np.array([i for i in range(150)])
    # Reshape values to 4D array (level, time, lats, lons)
    fvalues = fvalues.reshape(len(flevels), len(
        times), len(flatitudes), len(flongitudes))
    # Ingest values to netCDF file
    latitudes[:] = flatitudes
    longitudes[:] = flongitudes
    times[:] = ftimes
    levels[:] = flevels
    values[:] = fvalues
    # Assign time info to time variable
    netcdf_file.variables['time'].units = 'months since 2001-01-01 00:00:00'
    netcdf_file.close()
    return file_path


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
import datetime
import numpy as np
import pickle
import os
import ocw.data_source.rcmed as rcmed


class CustomAssertions:
    # Custom Assertions to handle Numpy Arrays

    def assert1DArraysEqual(self, array1, array2):
        self.assertSequenceEqual(tuple(array1), tuple(array2))


class test_rcmed(unittest.TestCase, CustomAssertions):

    def setUp(self):
        self.dataset_id = 2
        self.parameter_id = 15
        self.min_lat = 50
        self.max_lat = 70
        self.min_lon = 1
        self.max_lon = 15
        self.start_time = datetime.datetime(2002, 8, 1)
        self.end_time = datetime.datetime(2002, 10, 1)
        # start and end time for URL to query database is the beginning and end
        # of start_time and end_time
        self.start_time_for_url = "20020801T0000Z"
        self.end_time_for_url = "20021001T2359Z"
        self.url = b"http://rcmes.jpl.nasa.gov/query-api/query.php?"
        self.lats = np.arange(50.5, 70, 1)
        self.lons = np.arange(1.5, 15, 1)
        # In this parameter, two days of 10/20 and 10/21 have been missed.
        self.times_list = [datetime.datetime(
            2002, 8, 31) + datetime.timedelta(days=x) for x in range(0, 62)]
        self.times_list.remove(datetime.datetime(2002, 10, 20))
        self.times_list.remove(datetime.datetime(2002, 10, 21))
        self.times = np.array(self.times_list)

        # Get the full path to test_rcmed
        self.file_path = os.path.dirname(__file__)

        # Grab the parameter values file
        self.values = np.ma.array(np.load(os.path.join(
            self.file_path, "parameter_values.npy")))

        # Grab the parameter metadata file
        meta_file = open(os.path.join(
            self.file_path, "parameters_metadata_output.p"), "rb")
        self.param_metadata_output = pickle.load(meta_file)

    def return_text(self, url):
        url = _force_bytes(url)
        query = _force_bytes(("datasetId={0}&"
                              "parameterId={1}&"
                              "latMin={2}&"
                              "latMax={3}&"
                              "lonMin={4}&"
                              "lonMax={5}&"
                              "timeStart={6}&"
                              "timeEnd={7}").format(self.dataset_id,
                                                    self.parameter_id,
                                                    self.min_lat,
                                                    self.max_lat,
                                                    self.min_lon,
                                                    self.max_lon,
                                                    self.start_time_for_url,
                                                    self.end_time_for_url))
        if url == self.url + query:
            return open(os.path.join(self.file_path,
                                     "parameter_dataset_text.txt"), 'rb')
        elif url == self.url + b"&param_info=yes":
            return open(os.path.join(self.file_path,
                                     "parameters_metadata_text.txt"), 'rb')
        else:
            raise Exception("The URL is not acceptable.")

    def test_function_get_parameters_metadata(self):
        rcmed.urlopen = self.return_text
        self.assertEqual(rcmed.get_parameters_metadata(),
                         self.param_metadata_output)

    def test_function_parameter_dataset_lats_monthly(self):
        self.dataset_id = 6
        self.parameter_id = 32
        self.end_time_for_url = "20021031T0000Z"
        rcmed.urlopen = self.return_text
        self.assert1DArraysEqual(rcmed.parameter_dataset(self.dataset_id,
                                                         self.parameter_id,
                                                         self.min_lat,
                                                         self.max_lat,
                                                         self.min_lon,
                                                         self.max_lon,
                                                         self.start_time,
                                                         self.end_time).lats,
                                 self.lats)

    def test_function_parameter_dataset_lats(self):
        rcmed.urlopen = self.return_text
        self.assert1DArraysEqual(rcmed.parameter_dataset(self.dataset_id,
                                                         self.parameter_id,
                                                         self.min_lat,
                                                         self.max_lat,
                                                         self.min_lon,
                                                         self.max_lon,
                                                         self.start_time,
                                                         self.end_time).lats,
                                 self.lats)

    def test_function_parameter_dataset_lons(self):
        rcmed.urlopen = self.return_text
        self.assert1DArraysEqual(rcmed.parameter_dataset(self.dataset_id,
                                                         self.parameter_id,
                                                         self.min_lat,
                                                         self.max_lat,
                                                         self.min_lon,
                                                         self.max_lon,
                                                         self.start_time,
                                                         self.end_time).lons,
                                 self.lons)

    def test_function_parameter_dataset_times(self):
        rcmed.urlopen = self.return_text
        self.assert1DArraysEqual(rcmed.parameter_dataset(self.dataset_id,
                                                         self.parameter_id,
                                                         self.min_lat,
                                                         self.max_lat,
                                                         self.min_lon,
                                                         self.max_lon,
                                                         self.start_time,
                                                         self.end_time).times,
                                 self.times)

    def test_function_parameter_dataset_values(self):
        rcmed.urlopen = self.return_text
        self.assert1DArraysEqual(rcmed.parameter_dataset(self.dataset_id,
                                                         self.parameter_id,
                                                         self.min_lat,
                                                         self.max_lat,
                                                         self.min_lon,
                                                         self.max_lon,
                                                         self.start_time,
                                                         self.end_time)
                                 .values.flatten(),
                                 self.values.flatten())

    def test_function_parameter_dataset_name(self):
        rcmed.urlopen = self.return_text
        ds = rcmed.parameter_dataset(self.dataset_id,
                                     self.parameter_id,
                                     self.min_lat,
                                     self.max_lat,
                                     self.min_lon,
                                     self.max_lon,
                                     self.start_time,
                                     self.end_time,
                                     name='foo')
        self.assertEquals(ds.name, 'foo')

    def test_dataset_origin(self):
        rcmed.urlopen = self.return_text
        ds = rcmed.parameter_dataset(self.dataset_id,
                                     self.parameter_id,
                                     self.min_lat,
                                     self.max_lat,
                                     self.min_lon,
                                     self.max_lon,
                                     self.start_time,
                                     self.end_time,
                                     name='foo')

        self.assertEquals(ds.origin['source'], 'rcmed')
        self.assertEquals(ds.origin['dataset_id'], self.dataset_id)
        self.assertEquals(ds.origin['parameter_id'], self.parameter_id)


def _force_bytes(s, encoding='utf-8'):
    if hasattr(s, 'encode'):
        s = s.encode(encoding=encoding)
    
    return s

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''
Classes:
    Metric - Abstract Base Class from which all metrics must inherit.
'''

from abc import ABCMeta, abstractmethod
import ocw.utils as utils
import numpy
import numpy.ma as ma
from scipy.stats import mstats


class Metric(object):
    '''Base Metric Class'''
    __metaclass__ = ABCMeta


class UnaryMetric(Metric):
    '''Abstract Base Class from which all unary metrics inherit.'''
    __metaclass__ = ABCMeta

    @abstractmethod
    def run(self, target_dataset):
        '''Run the metric for a given target dataset.

        :param target_dataset: The dataset on which the current metric will
            be run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The result of evaluating the metric on the target_dataset.
        '''


class BinaryMetric(Metric):
    '''Abstract Base Class from which all binary metrics inherit.'''
    __metaclass__ = ABCMeta

    @abstractmethod
    def run(self, ref_dataset, target_dataset):
        '''Run the metric for the given reference and target datasets.

        :param ref_dataset: The Dataset to use as the reference dataset when
            running the evaluation.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The Dataset to use as the target dataset when
            running the evaluation.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The result of evaluation the metric on the reference and 
            target dataset.
        '''


class Bias(BinaryMetric):
    '''Calculate the bias between a reference and target dataset.'''

    def run(self, ref_dataset, target_dataset):
        '''Calculate the bias between a reference and target dataset.

        .. note::
           Overrides BinaryMetric.run()

        :param ref_dataset: The reference dataset to use in this metric run.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The difference between the reference and target datasets.
        :rtype: :class:`numpy.ndarray`
        '''
        return calc_bias(target_dataset.values, ref_dataset.values)


class SpatialPatternTaylorDiagram(BinaryMetric):
    ''' Calculate the target to reference ratio of spatial standard deviation and pattern correlation'''

    def run(self, ref_dataset, target_dataset):
        '''Calculate two metrics to plot a Taylor diagram to compare spatial patterns      

        .. note::
           Overrides BinaryMetric.run() 

        :param ref_dataset: The reference dataset to use in this metric run.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: standard deviation ratio, pattern correlation coefficient
        :rtype: :float:'float','float' 
        '''
        return ma.array([calc_stddev_ratio(target_dataset.values, ref_dataset.values), calc_correlation(target_dataset.values, ref_dataset.values)])


class TemporalStdDev(UnaryMetric):
    '''Calculate the standard deviation over the time.'''

    def run(self, target_dataset):
        '''Calculate the temporal std. dev. for a datasets.

        .. note::
           Overrides UnaryMetric.run()

        :param target_dataset: The target_dataset on which to calculate the 
            temporal standard deviation.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The temporal standard deviation of the target dataset
        :rtype: :class:`ndarray`
        '''
        return calc_stddev(target_dataset.values, axis=0)


class StdDevRatio(BinaryMetric):
    '''Calculate the standard deviation ratio between two datasets.'''

    def run(self, ref_dataset, target_dataset):
        '''Calculate the standard deviation ratio.

        .. note::
            Overrides BinaryMetric.run()

        :param ref_dataset: The reference dataset to use in this metric run.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The standard deviation ratio of the reference and target
        '''

        return calc_stddev_ratio(target_dataset.values, ref_dataset.values)


class PatternCorrelation(BinaryMetric):
    '''Calculate the correlation coefficient between two datasets'''

    def run(self, ref_dataset, target_dataset):
        '''Calculate the correlation coefficient between two dataset.

        .. note::
           Overrides BinaryMetric.run()

        :param ref_dataset: The reference dataset to use in this metric run.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The correlation coefficient between a reference and target dataset.
        '''
        # stats.pearsonr returns correlation_coefficient, 2-tailed p-value
        # We only care about the correlation coefficient
        # Docs at
        # http://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html

        return calc_correlation(target_dataset.values, ref_dataset.values)


class TemporalCorrelation(BinaryMetric):
    '''Calculate the temporal correlation coefficients and associated
       confidence levels between two datasets, using Pearson's correlation.'''

    def run(self, reference_dataset, target_dataset):
        '''Calculate the temporal correlation coefficients and associated
           confidence levels between two datasets, using Pearson's correlation.

        .. note::
           Overrides BinaryMetric.run()

        :param reference_dataset: The reference dataset to use in this metric
            run
        :type reference_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run
        :type target_dataset: :class:`dataset.Dataset`

        :returns: A 2D array of temporal correlation coefficients and a 2D
            array of confidence levels associated with the temporal correlation
            coefficients
        '''
        num_times, num_lats, num_lons = reference_dataset.values.shape
        coefficients = ma.zeros([num_lats, num_lons])
        for i in numpy.arange(num_lats):
            for j in numpy.arange(num_lons):
                coefficients[i, j] = calc_correlation(
                    target_dataset.values[:, i, j],
                    reference_dataset.values[:, i, j])
        return coefficients


class TemporalMeanBias(BinaryMetric):
    '''Calculate the bias averaged over time.'''

    def run(self, ref_dataset, target_dataset):
        '''Calculate the bias averaged over time.

        .. note::
           Overrides BinaryMetric.run()

        :param ref_dataset: The reference dataset to use in this metric run.
        :type ref_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run.
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The mean bias between a reference and target dataset over time.
        '''

        return calc_bias(target_dataset.values, ref_dataset.values, average_over_time=True)


class RMSError(BinaryMetric):
    '''Calculate the Root Mean Square Difference (RMS Error), with the mean
       calculated over time and space.'''

    def run(self, reference_dataset, target_dataset):
        '''Calculate the Root Mean Square Difference (RMS Error), with the mean
           calculated over time and space.

        .. note::
           Overrides BinaryMetric.run()

        :param reference_dataset: The reference dataset to use in this metric
            run
        :type reference_dataset: :class:`dataset.Dataset`

        :param target_dataset: The target dataset to evaluate against the
            reference dataset in this metric run
        :type target_dataset: :class:`dataset.Dataset`

        :returns: The RMS error, with the mean calculated over time and space
        '''

        return calc_rmse(target_dataset.values, reference_dataset.values)


def calc_bias(target_array, reference_array, average_over_time=False):
    ''' Calculate difference between two arrays

    :param target_array: an array to be evaluated, as model output
    :type target_array: :class:'numpy.ma.core.MaskedArray'

    :param reference_array: an array of reference dataset
    :type reference_array: :class:'numpy.ma.core.MaskedArray'

    :param average_over_time: if True, calculated bias is averaged for the axis=0
    :type average_over_time: 'bool'

    :returns: Biases array of the target dataset
    :rtype: :class:'numpy.ma.core.MaskedArray'
    '''

    bias = target_array - reference_array
    if average_over_time:
        return ma.average(bias, axis=0)
    else:
        return bias


def calc_stddev(array, axis=None):
    ''' Calculate a sample standard deviation of an array along the array

    :param array: an array to calculate sample standard deviation
    :type array: :class:'numpy.ma.core.MaskedArray'

    :param axis: Axis along which the sample standard deviation is computed.
    :type axis: 'int'

    :returns: sample standard deviation of array
    :rtype: :class:'numpy.ma.core.MaskedArray'
    '''

    if isinstance(axis, int):
        return ma.std(array, axis=axis, ddof=1)
    else:
        return ma.std(array, ddof=1)


def calc_stddev_ratio(target_array, reference_array):
    ''' Calculate ratio of standard deivations of the two arrays

    :param target_array: an array to be evaluated, as model output
    :type target_array: :class:'numpy.ma.core.MaskedArray'

    :param reference_array: an array of reference dataset
    :type reference_array: :class:'numpy.ma.core.MaskedArray'

    :param average_over_time: if True, calculated bias is averaged for the axis=0
    :type average_over_time: 'bool'

    :returns: (standard deviation of target_array)/(standard deviation of reference array)
    :rtype: :class:'float'
    '''

    return calc_stddev(target_array) / calc_stddev(reference_array)


def calc_correlation(target_array, reference_array):
    '''Calculate the correlation coefficient between two arrays.

    :param target_array: an array to be evaluated, as model output
    :type target_array: :class:'numpy.ma.core.MaskedArray'

    :param reference_array: an array of reference dataset
    :type reference_array: :class:'numpy.ma.core.MaskedArray'

    :returns: pearson's correlation coefficient between the two input arrays
    :rtype: :class:'numpy.ma.core.MaskedArray'
    '''

    return mstats.pearsonr(reference_array.flatten(), target_array.flatten())[0]


def calc_rmse(target_array, reference_array):
    ''' Calculate ratio of standard deivations of the two arrays

    :param target_array: an array to be evaluated, as model output
    :type target_array: :class:'numpy.ma.core.MaskedArray'

    :param reference_array: an array of reference dataset
    :type reference_array: :class:'numpy.ma.core.MaskedArray'

    :param average_over_time: if True, calculated bias is averaged for the axis=0
    :type average_over_time: 'bool'

    :returns: root mean square error
    :rtype: :class:'float'
    '''

    return (ma.mean((calc_bias(target_array, reference_array))**2))**0.5


def calc_histogram_overlap(hist1, hist2):
    ''' from Lee et al. (2014)
    :param hist1: a histogram array
    :type hist1: :class:'numpy.ndarray'
    :param hist2: a histogram array with the same size as hist1
    :type hist2: :class:'numpy.ndarray'
    '''

    hist1_flat = hist1.flatten()
    hist2_flat = hist2.flatten()

    if len(hist1_flat) != len(hist2_flat):
        err = "The two histograms have different sizes"
        raise ValueError(err)
    overlap = 0.
    for ii in numpy.arange(len(hist1_flat)):
        overlap = overlap + numpy.min([hist1_flat[ii], hist2_flat[ii]])
    return overlap


def calc_joint_histogram(data_array1, data_array2, bins_for_data1, bins_for_data2):
    ''' Calculate a joint histogram of two variables in data_array1 and data_array2
    :param data_array1: the first variable
    :type data_array1: :class:'numpy.ma.core.MaskedArray'
    :param data_array2: the second variable
    :type data_array2: :class:'numpy.ma.core.MaskedArray'
    :param bins_for_data1: histogram bin edges for data_array1
    :type bins_for_data1: :class:'numpy.ndarray'
    :param bins_for_data2: histogram bin edges for data_array2
    :type bins_for_data2: :class:'numpy.ndarray'
    '''
    if ma.count_masked(data_array1) != 0 or ma.count_masked(data_array2) != 0:
        index = numpy.where((data_array1.mask == False) &
                            (data_array2.mask == False))
        new_array1 = data_array1[index]
        new_array2 = data_array2[index]
    else:
        new_array1 = data_array1.flatten()
        new_array2 = data_array2.flatten()

    histo2d, xedge, yedge = numpy.histogram2d(new_array1, new_array2, bins=[
                                              bins_for_data1, bins_for_data2])
    return histo2d


def wet_spell_analysis(reference_array, threshold=0.1, nyear=1, dt=3.):
    ''' Characterize wet spells using sub-daily (hourly) data

    :param reference_array: an array to be analyzed
    :type reference_array: :class:'numpy.ma.core.MaskedArray'

    :param threshold: the minimum amount of rainfall [mm/hour] 
    :type threshold: 'float'

    :param nyear: the number of discontinous periods 
    :type nyear: 'int'

    :param dt: the temporal resolution of reference_array
    :type dt: 'float'
    '''
    nt = reference_array.shape[0]
    if reference_array.ndim == 3:
        reshaped_array = reference_array.reshape([nt, reference_array.size / nt])
    else:
        reshaped_array = reference_array
    if ma.count_masked(reshaped_array[0,:]) != 0:
        xy_indices = numpy.where(reshaped_array.mask[0, :] == False)[0]
    else:
        xy_indices = numpy.arange(reshaped_array.shape[1])

    nt_each_year = nt / nyear
    spell_duration = []
    peak_rainfall = []
    total_rainfall = []

    for index in xy_indices:
        for iyear in numpy.arange(nyear):
            data0_temp = reshaped_array[nt_each_year * iyear:nt_each_year * (iyear + 1),
                                        index]
            # time indices when precipitation rate is smaller than the
            # threshold [mm/hr]
            t_index = numpy.where((data0_temp <= threshold) &
                               (data0_temp.mask == False))[0]
            t_index = numpy.insert(t_index, 0, 0)
            t_index = t_index + nt_each_year * iyear
            for it in numpy.arange(t_index.size - 1):
                if t_index[it + 1] - t_index[it] > 1:
                    data1_temp = data0_temp[t_index[it] + 1:t_index[it + 1]]
                    if not ma.is_masked(data1_temp):
                        spell_duration.append(
                            (t_index[it + 1] - t_index[it] - 1) * dt)
                        peak_rainfall.append(data1_temp.max())
                        total_rainfall.append(data1_temp.sum())
    return numpy.array(spell_duration), numpy.array(peak_rainfall), numpy.array(total_rainfall)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''
Classes:
    DatasetLoader - Generate OCW Dataset objects from a variety of sources.
'''

import ocw.data_source.local as local
import ocw.data_source.rcmed as rcmed
import ocw.data_source.podaac_datasource as podaac
import warnings

class DatasetLoader:
    '''Generate a list of OCW Dataset objects from a variety of sources.'''

    def __init__(self, *loader_opts):
        '''Generate a list of OCW Dataset objects from a variety of sources.

        Each keyword argument can be information for a dataset in dictionary
        form. For example:

        >>> loader_opt1 = {'loader_name': 'rcmed', 'name': 'cru',
                           'dataset_id': 10, 'parameter_id': 34}
        >>> loader_opt2 = {'path': './data/TRMM_v7_3B43_1980-2010.nc,
                           'variable': 'pcp'}
        >>> loader = DatasetLoader(loader_opt1, loader_opt2)

        Or more conveniently if the loader configuration is defined in a
        yaml file named config_file (see RCMES examples):

        >>> import yaml
        >>> config = yaml.load(open(config_file))
        >>> obs_loader_config = config['datasets']['reference']
        >>> loader = DatasetLoader(*obs_loader_config)

        As shown in the first example, the dictionary for each argument should
        contain a loader name and parameters specific to the particular loader.
        Once the configuration is entered, the datasets may be loaded using:

        >>> loader.load_datasets()
        >>> obs_datasets = loader.datasets

        Additionally, each dataset must have a ``loader_name`` keyword. This may
        be one of the following:

        * ``'local'`` - One or multiple dataset files in a local directory
        * ``'local_split'`` - A single dataset split accross multiple files in a
                              local directory
        * ``'esgf'`` - Download the dataset from the Earth System Grid
                       Federation
        * ``'rcmed'`` - Download the dataset from the Regional Climate Model
                        Evaluation System Database
        * ``'dap'`` - Download the dataset from an OPeNDAP URL
        * ``'podaac'`` - Download the dataset from Physical Oceanography
                        Distributed Active Archive Center

        Users who wish to load datasets from loaders not described above may
        define their own custom dataset loader function and incorporate it as
        follows:

        >>> loader.add_source_loader('my_loader_name', my_loader_func)

        :param loader_opts: Dictionaries containing the each dataset loader
                            configuration, representing the keyword arguments of
                            the loader function specified by an additional key
                            called 'loader_name'. If not specified by the user,
                            this defaults to local.
        :type loader_opts: :class:`dict`

        :raises KeyError: If an invalid argument is passed to a data source
                            loader function.
        '''
        # dataset loader config
        self.set_loader_opts(*loader_opts)

        # Default loaders
        self._source_loaders = {
            'local': local.load_multiple_files,
            'local_split': local.load_dataset_from_multiple_netcdf_files,
            'rcmed': rcmed.parameter_dataset,
            'podaac': podaac.extract_l4_granule
        }
        
        # Exclude esgf and dap for python 3 until they are compatible
        try:
            import ocw.data_source.esgf as esgf
            import ocw.data_source.dap as dap
            self._source_loaders['dap'] = dap.load
            self._source_loaders['esgf'] = esgf.load_dataset
        except ImportError:
            warnings.warn('dap and esgf loaders missing. If these are needed, '
                          'fallback to python 2.7.x.')

    def add_source_loader(self, loader_name, loader_func):
        '''
        Add a custom source loader.

        :param loader_name: The name of the data source.
        :type loader_name: :mod:`string`

        :param loader_func: Reference to a custom defined function. This should
                            return an OCW Dataset object, and have an origin which satisfies
                            origin['source'] == loader_name.
        :type loader_func: :class:`callable`
        '''
        self._source_loaders[loader_name] = loader_func

    def add_loader_opts(self, *loader_opts):
        '''
        A convenient means of adding loader options for each dataset to the
        loader.

        :param loader_opts: Dictionaries containing the each dataset loader
                            configuration, representing the keyword arguments of
                            the loader function specified by an additional key
                            called 'loader_name'. If not specified by the user,
                            this defaults to local.
        :type loader_opts: :mod:`dict`
        '''
        for opt in loader_opts:
            if 'loader_name' not in opt:
                opt['loader_name'] = 'local'
        self._config.extend(loader_opts)

    def set_loader_opts(self, *loader_opts):
        '''
        Reset the dataset loader config.

        :param loader_opts: Dictionaries containing the each dataset loader
                            configuration, representing the keyword arguments of
                            the loader function specified by an additional key
                            called 'loader_name'. If not specified by the user,
                            this defaults to local.
        :type loader_opts: :mod:`dict`
        '''
        self._config = []
        self.add_loader_opts(*loader_opts)

    def load_datasets(self):
        '''
        Loads the datasets from the given loader configurations.
        '''
        # Ensure output is clear if loading is performed more than once to
        # prevent duplicates.
        self.datasets = []

        # Load the datasets
        for loader_opt in self._config:
            output = self._load(**loader_opt)

            # Need to account for the fact that some loaders return lists
            # of OCW Dataset objects instead of just one
            if isinstance(output, list):
                self.datasets.extend(output)
            else:
                self.datasets.append(output)

    def _load(self, **kwargs):
        '''
        Generic dataset loading method.
        '''
        # Extract the loader name
        loader_name = kwargs.pop('loader_name')

        # Find the correct loader function for the given data source
        loader_func = self._source_loaders[loader_name]

        # The remaining kwargs should be specific to the loader
        output = loader_func(**kwargs)

        # Preserve loader_name info for later use
        kwargs['loader_name'] = loader_name
        return output



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import re
import sys

import ocw.metrics as metrics

import yaml

logging.basicConfig()
logger = logging.getLogger(__name__)

def is_config_valid(config_data):
    """ Validate supplied evaluation configuration data.

    :param config_data: Dictionary of the data parsed from the supplied YAML
        configuration file.
    :type config_data: :func:`dict`

    :returns: True if the configuration data is sufficient for an evaluation and
        seems to be well formed, False otherwise.
    """
    if not _valid_minimal_config(config_data):
        logger.error('Insufficient configuration file data for an evaluation')
        return False

    if not _config_is_well_formed(config_data):
        logger.error('Configuration data is not well formed')
        return False

    return True

def _valid_minimal_config(config_data):
    """"""
    if not 'datasets' in config_data.keys():
        logger.error('No datasets specified in configuration data.')
        return False

    if not 'metrics' in config_data.keys():
        logger.error('No metrics specified in configuration data.')
        return False

    if _contains_unary_metrics(config_data['metrics']):
        if (not 'reference' in config_data['datasets'].keys() and 
            not 'targets' in config_data['datasets'].keys()):
            err = (
                'Unary metric in configuration data requires either a reference '
                'or target dataset to be present for evaluation. Please ensure '
                'that your config is well formed.'
            )
            logger.error(err)
            return False

    if _contains_binary_metrics(config_data['metrics']):
        if (not 'reference' in config_data['datasets'].keys() or 
            not 'targets' in config_data['datasets'].keys()):
            logger.error(
                'Binary metric in configuration requires both a reference '
                'and target dataset to be present for evaluation. Please ensure '
                'that your config is well formed.'
            )
            return False

    return True

def _config_is_well_formed(config_data):
    """"""
    is_well_formed = True

    if 'reference' in config_data['datasets']:
        if not _valid_dataset_config_data(config_data['datasets']['reference']):
            is_well_formed = False

    if 'targets' in config_data['datasets']:
        targets = config_data['datasets']['targets']
        if type(targets) != type(list()):
            err = (
                'Expected to find list of target datasets but instead found '
                'object of type {}'
            ).format(type(targets))
            logger.error(err)
            is_well_formed = False
        else:
            for t in targets:
                if not _valid_dataset_config_data(t):
                    is_well_formed = False

    available_metrics = _fetch_built_in_metrics()
    for metric in config_data['metrics']:
        if metric not in available_metrics:
            warning = (
                'Unable to locate metric name {} in built-in metrics. If this '
                'is not a user defined metric then please check for potential '
                'misspellings.'
            ).format(metric)
            logger.warn(warning)
            is_well_formed = False

    if 'subregions' in config_data:
        for subregion in config_data['subregions']:
            if not _valid_subregion_config_data(subregion):
                is_well_formed = False

    if 'plots' in config_data:
        for plot in config_data['plots']:
            if not _valid_plot_config_data(plot):
                is_well_formed = False
            # Ensure that if we're trying to make a plot that require
            # subregion info that the config has this present.
            elif plot['type'] in ['subregion', 'portrait']:
                if ('subregions' not in config_data or
                    len(config_data['subregions']) < 1):
                    logger.error(
                        'Plot config that requires subregion information is present '
                        'in a config file without adequate subregion information '
                        'provided. Please ensure that you have properly supplied 1 or '
                        'more subregion config values.'
                    )
                    is_well_formed = False


    return is_well_formed

def _contains_unary_metrics(config_metric_data):
    """"""
    unarys = [cls.__name__ for cls in metrics.UnaryMetric.__subclasses__()]
    return any(metric in unarys for metric in config_metric_data)

def _contains_binary_metrics(config_metric_data):
    """"""
    binarys = [cls.__name__ for cls in metrics.BinaryMetric.__subclasses__()]
    return any(metric in binarys for metric in config_metric_data)

def _fetch_built_in_metrics():
    """"""
    unarys = [cls.__name__ for cls in metrics.UnaryMetric.__subclasses__()]
    binarys = [cls.__name__ for cls in metrics.BinaryMetric.__subclasses__()]
    return unarys + binarys

def _valid_dataset_config_data(dataset_config_data):
    """"""
    try:
        data_source = dataset_config_data['data_source']
    except KeyError:
        logger.error('Dataset does not contain a data_source attribute.')
        return False

    if data_source == 'local':
        required_keys = set(['data_source', 'file_count', 'path', 'variable'])
    elif data_source == 'rcmed':
        required_keys = set([
            'dataset_id',
            'parameter_id',
            'min_lat',
            'max_lat',
            'min_lon',
            'max_lon',
            'start_time',
            'end_time',
        ])
    elif data_source == 'esgf':
        required_keys = set([
            'data_source',
            'dataset_id',
            'variable',
            'esgf_username',
            'esgf_password'
        ])
    elif data_source == 'dap':
        required_keys = set({'url', 'variable'})
    else:
        logger.error('Dataset does not contain a valid data_source location.')
        return False

    present_keys = set(dataset_config_data.keys())
    missing_keys = required_keys - present_keys
    contains_required = len(missing_keys) == 0

    if contains_required:
        if data_source == 'local' and dataset_config_data['file_count'] > 1:
            # If the dataset is a multi-file dataset then we need to make sure
            # that the file glob pattern is included.
            if not 'file_glob_pattern' in dataset_config_data:
                logger.error(
                    'Multi-file local dataset is missing key: file_glob_pattern'
                )
                return False
        return True
    else:
        missing = sorted(list(missing_keys))
        logger.error(
            'Dataset does not contain required keys. '
            'The following keys are missing: {}'.format(', '.join(missing))
        )
        return False

def _valid_plot_config_data(plot_config_data):
    """"""
    try:
        plot_type = plot_config_data['type']
    except KeyError:
        logger.error('Plot config does not include a type attribute.')
        return False

    if plot_type == 'contour':
        required_keys = set([
            'results_indices',
            'lats',
            'lons',
            'output_name'
        ])
    elif plot_type == 'taylor':
        required_keys = set([
            'stddev_results_indices',
            'pattern_corr_results_indices',
            'output_name'
        ])
    elif plot_type == 'subregion':
        required_keys = set([
            'lats',
            'lons',
            'output_name'
        ])
    elif plot_type == 'time_series':
        required_keys = set([
            'temporal_boundaries'
        ])
    elif plot_type == 'portrait':
        required_keys = set([
            'metric_index',
            'output_name'
        ])
    else:
        logger.error('Invalid plot type specified.')
        return False

    present_keys = set(plot_config_data.keys())
    missing_keys = required_keys - present_keys
    contains_required = len(missing_keys) == 0

    if not contains_required:
        missing = sorted(list(missing_keys))
        logger.error(
            'Plot config does not contain required keys. '
            'The following keys are missing: {}'.format(', '.join(missing))
        )
        return False

    return True

def _valid_subregion_config_data(subregion_config_data):
    """"""
    if type(subregion_config_data) != type([]):
        logger.error(
            'Subregions should be passed as a list of lists where '
            'each sub-list contains a bounding box of the form: '
            '[lat_min, lat_max, lon_min, lon_max].'
        )
        return False

    if len(subregion_config_data) != 4:
        logger.error(
            'Subregions should be passed as a list of lists where '
            'each sub-list contains a bounding box of the form: '
            '[lat_min, lat_max, lon_min, lon_max].'
        )
        return False

    return True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
import os
import datetime
from dateutil.relativedelta import relativedelta

import netCDF4
import numpy as np

from ocw.dataset import Dataset
import ocw.utils as utils


class TestDecodeTimes(unittest.TestCase):

    def setUp(self):
        path = os.path.dirname(os.path.realpath(__file__))
        self.test_model = path + \
            '/../../ocw-ui/backend/tests/example_data/lat_lon_time.nc'
        self.test_model_daily = path + '/lat_lon_time_daily.nc'

        self.netcdf = netCDF4.Dataset(
            os.path.abspath(self.test_model), mode='r')
        self.netcdf_daily = netCDF4.Dataset(
            os.path.abspath(self.test_model_daily), mode='r')

    def test_proper_return_format(self):
        times = utils.decode_time_values(self.netcdf, 'time')

        self.assertTrue(all([type(x) is datetime.datetime for x in times]))

    def test_valid_time_processing(self):
        start_time = datetime.datetime.strptime(
            '1989-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')
        end_time = datetime.datetime.strptime(
            '2008-12-01 00:00:00', '%Y-%m-%d %H:%M:%S')
        times = utils.decode_time_values(self.netcdf, 'time')
        self.assertEquals(times[0], start_time)
        self.assertEquals(times[-1], end_time)

    def test_days_time_processing(self):
        start_time = datetime.datetime.strptime(
            '1951-4-14 00:00:00', '%Y-%m-%d %H:%M:%S')
        end_time = datetime.datetime.strptime(
            '1951-12-9 00:00:00', '%Y-%m-%d %H:%M:%S')
        new_times = utils.decode_time_values(self.netcdf_daily, 'time')
        self.assertEqual(new_times[0], start_time)
        self.assertEqual(new_times[-1], end_time)


class TestTimeUnitsParse(unittest.TestCase):

    def test_valid_parse(self):
        units = utils.parse_time_units('minutes since a made up date')

        self.assertEquals(units, 'minutes')

    def test_invalid_parse(self):
        self.assertRaises(
            ValueError,
            utils.parse_time_units,
            'parsecs since a made up date'
        )


class TestTimeBaseParse(unittest.TestCase):

    def test_valid_time_base(self):
        base_time = utils.parse_time_base('days since 1988-06-10 00:00:00')
        start_time = datetime.datetime.strptime(
            '1988-06-10 00:00:00', '%Y-%m-%d %H:%M:%S')

        self.assertEquals(base_time, start_time)

    def test_invalid_time_base(self):
        self.assertRaises(
            ValueError,
            utils.parse_time_base,
            'days since 1988g06g10g00g00g00'
        )


class TestBaseTimeStringParse(unittest.TestCase):

    def test_valid_time_base_string_parse(self):
        base = utils.parse_base_time_string('days since 1988-06-10 00:00:00')

        self.assertEquals(base, '1988-06-10 00:00:00')

    def test_invalid_time_base_string_parse(self):
        self.assertRaises(
            ValueError,
            utils.parse_base_time_string,
            'this string is not valid'
        )


class TestNormalizeLatLonValues(unittest.TestCase):

    def setUp(self):
        times = np.array([datetime.datetime(2000, x, 1) for x in range(1, 13)])
        self.lats = np.arange(-30, 30)
        self.lons = np.arange(360)
        flat_array = np.arange(len(times) * len(self.lats) * len(self.lons))
        self.values = flat_array.reshape(
            len(times), len(self.lats), len(self.lons))
        self.lats2 = np.array([-30, 0, 30])
        self.lons2 = np.array([0, 100, 200, 300])
        self.values2 = np.arange(12).reshape(3, 4)
        self.lats_unsorted = np.array([-30, 20, -50])
        self.lons_unsorted = np.array([-30, 20, -50, 40])

    def test_full_lons_shift(self):
        lats, lons, values = utils.normalize_lat_lon_values(self.lats,
                                                            self.lons,
                                                            self.values)
        np.testing.assert_array_equal(lons, np.arange(-180, 180))

    def test_lats_reversed(self):
        lons2 = np.arange(-180, 180)
        lats, lons, values = utils.normalize_lat_lon_values(self.lats[::-1],
                                                            lons2,
                                                            self.values[:,
                                                                        ::-1,
                                                                        :])
        np.testing.assert_array_equal(lats, self.lats)
        np.testing.assert_array_equal(values, self.values)

    def test_lons_reversed(self):
        self.lats = np.arange(-10, 10)
        self.lons = np.arange(40)
        times = np.array([datetime.datetime(2000, x, 1) for x in range(1, 7)])
        flat_array = np.arange(len(times) * len(self.lats) * len(self.lons))
        self.variable = flat_array.reshape(len(times),
                                           len(self.lats),
                                           len(self.lons))
        lats, lons, values = utils.normalize_lat_lon_values(self.lats,
                                                            self.lons[::-1],
                                                            self.values[:,
                                                                        ::,
                                                                        ::-1])
        np.testing.assert_array_equal(lats, self.lats)
        np.testing.assert_array_equal(values, self.values)
        np.testing.assert_array_equal(lons, self.lons)

    def test_lons_shift_values(self):
        expected_vals = np.array([[2, 3, 0, 1],
                                  [6, 7, 4, 5],
                                  [10, 11, 8, 9]])
        lats, lons, values = utils.normalize_lat_lon_values(self.lats2,
                                                            self.lons2,
                                                            self.values2)
        np.testing.assert_array_equal(values, expected_vals)

    def test_shift_and_reversed(self):
        expected_vals = np.array([[10, 11, 8, 9],
                                  [6, 7, 4, 5],
                                  [2, 3, 0, 1]])
        lats, lons, values = utils.normalize_lat_lon_values(self.lats2[::-1],
                                                            self.lons2,
                                                            self.values2)
        np.testing.assert_array_equal(values, expected_vals)

    def test_lats_not_sorted(self):
        self.assertRaises(ValueError,
                          utils.normalize_lat_lon_values,
                          self.lats_unsorted,
                          self.lons2,
                          self.values2)

    def test_lons_not_sorted(self):
        self.assertRaises(ValueError,
                          utils.normalize_lat_lon_values,
                          self.lats2,
                          self.lons_unsorted,
                          self.values2)

    def test_lons_greater_than_180(self):
        self.lons = np.array([190, 210, 230, 250])
        self.lats = np.array([-30, 0, 30])
        self.values = np.arange(12).reshape(3, 4)
        expected_lons = np.array([-170, -150, -130, -110])
        expected_values = np.array([[0, 1, 2, 3],
                                    [4, 5, 6, 7],
                                    [8, 9, 10, 11]])
        lats, lons, values = utils.normalize_lat_lon_values(self.lats,
                                                            self.lons,
                                                            self.values)
        np.testing.assert_array_equal(lons, expected_lons)
        np.testing.assert_array_equal(expected_values, values)


class TestGetTemporalOverlap(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array(
            [datetime.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.value = flat_array.reshape(12, 5, 5)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        self.dataset_array = [self.test_dataset, self.test_dataset]

    def test_same_dataset_temporal_overlap(self):
        maximum, minimum = utils.get_temporal_overlap(self.dataset_array)
        self.assertEqual(maximum, datetime.datetime(2000, 1, 1))
        self.assertEqual(minimum, datetime.datetime(2000, 12, 1))

    def test_different_dataset_temporal_overlap(self):
        new_times = np.array(
            [datetime.datetime(2002, x, 1) for x in range(1, 13)])
        another_dataset = Dataset(self.lat, self.lon, new_times,
                                  self.value, self.variable)
        self.dataset_array = [self.test_dataset, another_dataset]
        maximum, minimum = utils.get_temporal_overlap(self.dataset_array)
        self.assertEqual(maximum, datetime.datetime(2002, 1, 1))
        self.assertEqual(minimum, datetime.datetime(2000, 12, 1))


class TestReshapeMonthlyToAnnually(unittest.TestCase):
    ''' Testing function 'reshape_monthly_to_annually' from ocw.utils.py '''

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array(
            [datetime.datetime(2000, 1, 1) + relativedelta(months=x)
             for x in range(24)])
        flat_array = np.array(range(600))
        self.value = flat_array.reshape(24, 5, 5)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)

    def test_reshape_full_year(self):
        new_values = self.value.reshape(2, 12, 5, 5)
        np.testing.assert_array_equal(
            utils.reshape_monthly_to_annually(self.test_dataset), new_values)

    def test_reshape_not_full_year(self):
        new_time = np.array(
            [datetime.datetime(2000, 1, 1) + relativedelta(months=x)
             for x in range(26)])
        flat_array = np.array(range(650))
        value = flat_array.reshape(26, 5, 5)
        bad_dataset = Dataset(self.lat, self.lon,
                              new_time, value, self.variable)

        self.assertRaises(
            ValueError, utils.reshape_monthly_to_annually, bad_dataset)


class TestCalcTemporalMean(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14])
        self.lon = np.array([100, 102, 104])
        self.time = np.array(
            [datetime.datetime(2000, x, 1) for x in range(1, 7)])
        flat_array = np.array(range(54))
        self.value = flat_array.reshape(6, 3, 3)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)

    def test_returned_mean(self):
        mean_values = np.array([[22.5, 23.5, 24.5],
                                [25.5, 26.5, 27.5],
                                [28.5, 29.5, 30.5]])

        result = utils.calc_temporal_mean(self.test_dataset)
        np.testing.assert_array_equal(result, mean_values)


class TestCalcAreaWeightedSpatialAverage(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14])
        self.lon = np.array([100, 102, 104])
        self.time = np.array(
            [datetime.datetime(2000, x, 1) for x in range(1, 7)])
        flat_array = np.array(range(54))
        self.value = flat_array.reshape(6, 3, 3)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)

    def test_spatial_average(self):
        avg = np.ma.array([4., 13., 22., 31., 40., 49.])
        result = utils.calc_area_weighted_spatial_average(self.test_dataset)
        np.testing.assert_array_equal(avg, result)

    def test_2_dim_lats_lons(self):
        self.lat = np.array([10, 12, 14]).reshape(3, 1)
        self.lon = np.array([100, 102, 104]).reshape(3, 1)
        flat_array = np.array(range(18))
        self.value = flat_array.reshape(6, 3, 1)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        avg = np.ma.array([1., 4., 7., 10., 13., 16.])
        result = utils.calc_area_weighted_spatial_average(self.test_dataset)
        np.testing.assert_array_equal(avg, result)

    def test_spatial_average_with_area_weight(self):
        avg = np.ma.array([3.985158, 12.985158, 21.985158,
                           30.985158, 39.985158, 48.985158])
        result = utils.calc_area_weighted_spatial_average(
            self.test_dataset, area_weight=True)
        np.testing.assert_array_almost_equal(avg, result)

    def test__2_dim_lats_lons_area_weight(self):
        self.lat = np.array([10, 12, 14]).reshape(3, 1)
        self.lon = np.array([100, 102, 104]).reshape(3, 1)
        flat_array = np.array(range(18))
        self.value = flat_array.reshape(6, 3, 1)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        avg = np.ma.array([0.995053, 3.995053, 6.995053, 9.995053, 12.995053,
                           15.995053])
        result = utils.calc_area_weighted_spatial_average(
            self.test_dataset, area_weight=True)
        np.testing.assert_array_almost_equal(avg, result)


class TestCalcClimatologyYear(unittest.TestCase):
    ''' Testing function 'calc_climatology_year' from ocw.utils.py '''

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array(
            [datetime.datetime(2000, 1, 1) + relativedelta(months=x)
             for x in range(24)])
        flat_array = np.array(range(600))
        self.value = flat_array.reshape(24, 5, 5)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)

    def test_annually_mean(self):
        annually_mean = np.append(
            np.arange(137.5, 162.5, 1), np.arange(437.5, 462.5, 1))
        annually_mean.shape = (2, 5, 5)
        np.testing.assert_array_equal(
            utils.calc_climatology_year(self.test_dataset)[0], annually_mean)

    def test_total_mean(self):
        total_mean = np.arange(287.5, 312.5, 1)
        total_mean.shape = (5, 5)
        np.testing.assert_array_equal(
            utils.calc_climatology_year(self.test_dataset)[1], total_mean)

    def test_invalid_time_shape(self):
        flat_array = np.array(range(350))
        self.test_dataset.values = flat_array.reshape(14, 5, 5)
        with self.assertRaises(ValueError):
            utils.calc_climatology_year(self.test_dataset)


class TestCalcClimatologyMonthly(unittest.TestCase):
    ''' Tests the 'calc_climatology_monthly' method from ocw.utils.py '''

    def setUp(self):
        self.lats = np.array([10, 20, 30, 40, 50])
        self.lons = np.array([20, 30, 40, 50, 60])
        start_date = datetime.datetime(2000, 1, 1)
        self.times = np.array([start_date + relativedelta(months=x)
                               for x in range(36)])
        self.values = np.array([1] * 300 + [2] * 300 +
                               [0] * 300).reshape(36, 5, 5)
        self.variable = 'testdata'
        self.dataset = Dataset(self.lats, self.lons, self.times,
                               self.values, self.variable)

    def test_calc_climatology_monthly(self):
        expected_result = np.ones(300).reshape(12, 5, 5)
        expected_times = np.array([datetime.datetime(1, 1, 1) +
                                   relativedelta(months=x)
                                   for x in range(12)])
        actual_result, actual_times = utils.calc_climatology_monthly(
            self.dataset)
        np.testing.assert_array_equal(actual_result, expected_result)
        np.testing.assert_array_equal(actual_times, expected_times)

    def test_invalid_time_shape(self):
        flat_array = np.array(range(350))
        self.dataset.values = flat_array.reshape(14, 5, 5)
        with self.assertRaises(ValueError):
            utils.calc_climatology_monthly(self.dataset)


class TestCalcTimeSeries(unittest.TestCase):
    ''' Tests the 'calc_time_series' method from ocw.utils.py '''

    def setUp(self):
        self.lats = np.array([10, 20, 30, 40, 50])
        self.lons = np.array([20, 30, 40, 50, 60])
        start_date = datetime.datetime(2000, 1, 1)
        self.times = np.array([start_date + relativedelta(months=x)
                               for x in range(12)])
        self.values = np.ones(300).reshape(12, 5, 5)
        self.variable = 'testdata'
        self.dataset = Dataset(self.lats, self.lons, self.times,
                               self.values, self.variable)

    def test_calc_time_series(self):
        expected_result = np.ones(12)
        np.testing.assert_array_equal(
            utils.calc_time_series(self.dataset), expected_result)

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''Unit tests for the Evaluation.py module'''

import unittest
import numpy as np
import datetime as dt
from ocw.dataset import Dataset, Bounds
from ocw.evaluation import Evaluation
from ocw.metrics import Bias, TemporalStdDev


class TestEvaluation(unittest.TestCase):

    def setUp(self):
        self.eval = Evaluation(None, [], [])

        lat = np.array([10, 12, 14, 16, 18])
        lon = np.array([100, 102, 104, 106, 108])
        time = np.array([dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        value = flat_array.reshape(12, 5, 5)
        self.variable = 'prec'
        self.other_var = 'temp'
        self.test_dataset = Dataset(lat, lon, time, value, self.variable)
        self.another_test_dataset = Dataset(lat, lon, time, value,
                                            self.other_var)

    def test_init(self):
        self.assertEquals(self.eval.ref_dataset, None)
        self.assertEquals(self.eval.target_datasets, [])
        self.assertEquals(self.eval.metrics, [])
        self.assertEquals(self.eval.unary_metrics, [])

    def test_full_init(self):
        self.eval = Evaluation(
            self.test_dataset,
            [self.test_dataset, self.another_test_dataset],
            [Bias(), Bias(), TemporalStdDev()])
        ref_dataset = self.test_dataset
        target_datasets = [self.test_dataset, self.another_test_dataset]
        metrics = [Bias(), Bias()]
        unary_metrics = [TemporalStdDev()]

        self.eval = Evaluation(ref_dataset,
                               target_datasets,
                               metrics + unary_metrics)

        self.assertEqual(self.eval.ref_dataset.variable, self.variable)

        # Make sure the two target datasets were added properly
        self.assertEqual(self.eval.target_datasets[0].variable, self.variable)
        self.assertEqual(self.eval.target_datasets[1].variable, self.other_var)

        # Make sure the three metrics were added properly
        # The two Bias metrics are "binary" metrics
        self.assertEqual(len(self.eval.metrics), 2)
        # TemporalStdDev is a "unary" metric and should be stored as such
        self.assertEqual(len(self.eval.unary_metrics), 1)
        self.eval.run()
        out_str = (
            "<Evaluation - ref_dataset: {}, "
            "target_dataset(s): {}, "
            "binary_metric(s): {}, "
            "unary_metric(s): {}, "
            "subregion(s): {}>"
        ).format(
            str(self.test_dataset),
            [str(ds) for ds in target_datasets],
            [str(m) for m in metrics],
            [str(u) for u in unary_metrics],
            None
        )
        self.assertEqual(str(self.eval), out_str)

    def test_valid_ref_dataset_setter(self):
        self.eval.ref_dataset = self.another_test_dataset
        self.assertEqual(self.eval.ref_dataset.variable,
                         self.another_test_dataset.variable)

    def test_invalid_ref_dataset(self):
        with self.assertRaises(TypeError):
            self.eval.ref_dataset = "This isn't a Dataset object"

    def test_valid_subregion(self):
        bound = Bounds(
            lat_min=-10, lat_max=10,
            lon_min=-20, lon_max=20,
            start=dt.datetime(2000, 1, 1), end=dt.datetime(2001, 1, 1))

        self.eval.subregions = [bound, bound]
        self.assertEquals(len(self.eval.subregions), 2)

    def test_invalid_subregion_bound(self):
        bound = "This is not a bounds object"

        with self.assertRaises(TypeError):
            self.eval.subregions = [bound]

    def test_add_ref_dataset(self):
        self.eval = Evaluation(self.test_dataset, [], [])

        self.assertEqual(self.eval.ref_dataset.variable, self.variable)

    def test_add_valid_dataset(self):
        self.eval.add_dataset(self.test_dataset)

        self.assertEqual(self.eval.target_datasets[0].variable,
                         self.variable)

    def test_add_invalid_dataset(self):
        with self.assertRaises(TypeError):
            self.eval.add_dataset('This is an invalid dataset')

    def test_add_datasets(self):
        self.eval.add_datasets([self.test_dataset, self.another_test_dataset])

        self.assertEqual(len(self.eval.target_datasets), 2)
        self.assertEqual(self.eval.target_datasets[0].variable,
                         self.variable)
        self.assertEqual(self.eval.target_datasets[1].variable,
                         self.other_var)

    def test_add_valid_metric(self):
        # Add a "binary" metric
        self.assertEqual(len(self.eval.metrics), 0)
        self.eval.add_metric(Bias())
        self.assertEqual(len(self.eval.metrics), 1)

        # Add a "unary" metric
        self.assertEqual(len(self.eval.unary_metrics), 0)
        self.eval.add_metric(TemporalStdDev())
        self.assertEqual(len(self.eval.unary_metrics), 1)

    def test_add_invalid_metric(self):
        with self.assertRaises(TypeError):
            self.eval.add_metric('This is an invalid metric')

    def test_add_metrics(self):
        self.assertEqual(len(self.eval.metrics), 0)
        self.eval.add_metrics([Bias(), Bias()])
        self.assertEqual(len(self.eval.metrics), 2)

    def test_invalid_evaluation_run(self):
        self.eval = Evaluation(None, [], [])
        self.assertEqual(self.eval.run(), None)

    def test_bias_output_shape(self):
        bias_eval = Evaluation(self.test_dataset, [
                               self.another_test_dataset], [Bias()])
        bias_eval.run()
        input_shape = tuple(self.test_dataset.values.shape)
        bias_results_shape = tuple(bias_eval.results[0][0].shape)
        self.assertEqual(input_shape, bias_results_shape)

    def test_result_shape(self):
        bias_eval = Evaluation(
            self.test_dataset,
            [self.another_test_dataset, self.another_test_dataset,
                self.another_test_dataset],
            [Bias(), Bias()]
        )
        bias_eval.run()

        # Expected result shape is
        # [bias, bias] where bias.shape[0] = number of datasets
        self.assertTrue(len(bias_eval.results) == 2)
        self.assertTrue(bias_eval.results[0].shape[0] == 3)

    def test_unary_result_shape(self):
        new_eval = Evaluation(
            self.test_dataset,
            [self.another_test_dataset, self.another_test_dataset,
                self.another_test_dataset, self.another_test_dataset],
            [TemporalStdDev()]
        )
        new_eval.run()

        # Expected result shape is
        # [stddev] where stddev.shape[0] = number of datasets

        self.assertTrue(len(new_eval.unary_results) == 1)
        self.assertTrue(new_eval.unary_results[0].shape[0] == 5)

    def test_subregion_result_shape(self):
        bound = Bounds(
            lat_min=10, lat_max=18,
            lon_min=100, lon_max=108,
            start=dt.datetime(2000, 1, 1), end=dt.datetime(2000, 3, 1))

        bias_eval = Evaluation(
            self.test_dataset,
            [self.another_test_dataset, self.another_test_dataset],
            [Bias()],
            [bound]
        )
        bias_eval.run()

        # Expected result shape is
        # [
        #       [   # Subregions cause this extra layer
        #           [number of targets, bias.run(reference, target1).shape]
        #       ]
        #   ],
        self.assertTrue(len(bias_eval.results) == 1)
        self.assertTrue(len(bias_eval.results[0]) == 1)
        self.assertTrue(bias_eval.results[0][0].shape[0] == 2)
        self.assertTrue(isinstance(bias_eval.results, type([])))

    def test_subregion_unary_result_shape(self):
        bound = Bounds(
            lat_min=10, lat_max=18,
            lon_min=100, lon_max=108,
            start=dt.datetime(2000, 1, 1), end=dt.datetime(2000, 3, 1))

        new_eval = Evaluation(
            self.test_dataset,
            [self.another_test_dataset, self.another_test_dataset],
            [TemporalStdDev(), TemporalStdDev()],
            [bound, bound, bound, bound, bound]
        )
        new_eval.run()

        # Expected result shape is
        # [
        #       [   # Subregions cause this extra layer
        #           [3, temporalstddev.run(reference).shape],
        #       ]
        # ]

        # 5 = number of subregions
        self.assertTrue(len(new_eval.unary_results) == 5)
        # number of metrics
        self.assertTrue(len(new_eval.unary_results[0]) == 2)
        self.assertTrue(isinstance(new_eval.unary_results, type([])))
        # number of datasets (ref + target)
        self.assertTrue(new_eval.unary_results[0][0].shape[0] == 3)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''Unit tests for the Dataset.py module'''

import unittest
from ocw.dataset import Dataset, Bounds
import numpy as np
import datetime as dt


class TestDatasetAttributes(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array([dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.value = flat_array.reshape(12, 5, 5)
        self.variable = 'prec'
        self.name = 'foo'
        self.origin = {'path': '/a/fake/file/path'}
        self.test_dataset = Dataset(self.lat,
                                    self.lon,
                                    self.time,
                                    self.value,
                                    variable=self.variable,
                                    name=self.name,
                                    origin=self.origin)

    def test_lats(self):
        np.testing.assert_array_equal(self.test_dataset.lats, self.lat)

    def test_lons(self):
        np.testing.assert_array_equal(self.test_dataset.lons, self.lon)

    def test_times(self):
        np.testing.assert_array_equal(self.test_dataset.times, self.time)

    def test_values(self):
        self.assertEqual(self.test_dataset.values.all(), self.value.all())

    def test_variable(self):
        self.assertEqual(self.test_dataset.variable, self.variable)

    def test_name(self):
        self.assertEqual(self.test_dataset.name, self.name)

    def test_origin(self):
        self.assertEqual(self.test_dataset.origin, self.origin)


class TestInvalidDatasetInit(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array([dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.value = flat_array.reshape(12, 5, 5)
        self.values_in_wrong_order = flat_array.reshape(5, 5, 12)

    def test_bad_lat_shape(self):
        self.lat = np.array([[1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')

    def test_bad_lon_shape(self):
        self.lon = np.array([[1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')

    def test_bad_times_shape(self):
        self.time = np.array([[1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')

    def test_bad_values_shape(self):
        self.value = np.array([1, 2, 3, 4, 5])
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')
        self.value = self.value.reshape(1, 5)
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')

    def test_values_shape_mismatch(self):
        # If we change lats to this the shape of value will not match
        # up with the length of the lats array.
        self.lat = self.lat[:-2]
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.value, 'prec')

    def test_values_given_in_wrong_order(self):
        with self.assertRaises(ValueError):
            Dataset(self.lat, self.lon, self.time, self.values_in_wrong_order)

    def test_lons_values_incorrectly_gridded(self):
        times = np.array([dt.datetime(2000, x, 1) for x in range(1, 13)])
        lats = np.arange(-30, 30)
        bad_lons = np.arange(360)
        flat_array = np.arange(len(times) * len(lats) * len(bad_lons))
        values = flat_array.reshape(len(times), len(lats), len(bad_lons))

        ds = Dataset(lats, bad_lons, times, values)
        np.testing.assert_array_equal(ds.lons, np.arange(-180, 180))

    def test_reversed_lats(self):
        ds = Dataset(self.lat[::-1], self.lon, self.time, self.value)
        np.testing.assert_array_equal(ds.lats, self.lat)


class TestDatasetFunctions(unittest.TestCase):

    def setUp(self):
        self.lat = np.array([10, 12, 14, 16, 18])
        self.lon = np.array([100, 102, 104, 106, 108])
        self.time = np.array([dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.value = flat_array.reshape(12, 5, 5)
        self.variable = 'prec'
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)

    def test_spatial_boundaries(self):
        self.assertEqual(
            self.test_dataset.spatial_boundaries(),
            (min(self.lat), max(self.lat), min(self.lon), max(self.lon)))

    def test_temporal_boundaries(self):
        self.assertEqual(
            self.test_dataset.temporal_boundaries(),
            (dt.datetime(2000, 1, 1), dt.datetime(2000, 12, 1)))

    def test_spatial_resolution(self):
        self.assertEqual(self.test_dataset.spatial_resolution(), (2, 2))

    def test_spatial_resolution_2_dim_lat_lon(self):
        self.lat = np.array([10, 12, 14, 16, 18, 20])
        self.lon = np.array([100, 102, 104, 106, 108, 110])
        self.lat = self.lat.reshape(3, 2)
        self.lon = self.lon.reshape(3, 2)
        flat_array = np.array(range(72))
        self.value = flat_array.reshape(12, 3, 2)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        self.assertEqual(self.test_dataset.spatial_resolution(), (6, 6))

    def test_temporal_resolution_hourly(self):
        self.time = np.array([dt.datetime(2000, 1, 1),
                              dt.datetime(2000, 1, 1)])
        flat_array = np.array(range(50))
        self.value = flat_array.reshape(2, 5, 5)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        self.assertEqual(self.test_dataset.temporal_resolution(), 'minutely')

    def test_temporal_resolution_monthly(self):
        self.assertEqual(self.test_dataset.temporal_resolution(), 'monthly')

    def test_temporal_resolution_daily(self):
        self.time = np.array([dt.datetime(2000, 3, x) for x in range(1, 31)])
        flat_array = np.array(range(750))
        self.value = flat_array.reshape(30, 5, 5)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        self.assertEqual(self.test_dataset.temporal_resolution(), 'daily')

    def test_temporal_resolution_yearly(self):
        self.time = np.array([dt.datetime(x, 6, 1) for x in range(2000, 2015)])
        flat_array = np.array(range(375))
        self.value = flat_array.reshape(15, 5, 5)
        self.test_dataset = Dataset(self.lat, self.lon, self.time,
                                    self.value, self.variable)
        self.assertEqual(self.test_dataset.temporal_resolution(), 'yearly')

    def test_str_(self):
        dataset = self.test_dataset
        lat_min, lat_max, lon_min, lon_max = dataset.spatial_boundaries()
        start, end = dataset.temporal_boundaries()
        lat_range = "({}, {})".format(lat_min, lon_min)
        lon_range = "({}, {})".format(lon_min, lon_min)
        temporal_boundaries = "({}, {})".format(start, end)

        formatted_repr = (
            "<Dataset - name: {}, "
            "lat-range: {}, "
            "lon-range: {}, "
            "temporal_boundaries: {}, "
            "var: {}, "
            "units: {}>"
        )

        output = formatted_repr.format(
            dataset.name if dataset.name != "" else None,
            lat_range,
            lon_range,
            temporal_boundaries,
            dataset.variable,
            dataset.units
        )
        self.assertEqual(str(self.test_dataset), output)


class TestBounds(unittest.TestCase):

    def setUp(self):
        self.bounds_rectangular = Bounds(lat_min=-80, lat_max=80,                # Lats
                                         lon_min=-160, lon_max=160,               # Lons
                                         start=dt.datetime(
                                             2000, 1, 1),  # Start time
                                         end=dt.datetime(2002, 1, 1))  # End time
        self.bounds_CORDEX = Bounds(boundary_type='CORDEX South Asia')
        self.bounds_us_states = Bounds(
            boundary_type='us_states', us_states=['CA', 'NV', 'AZ'])
        self.bounds_countries = Bounds(boundary_type='countries', countries=[
                                       'United States', 'Canada', 'Mexico'])

    def test_keywords(self):
        self.assertEqual(self.bounds_rectangular.boundary_type, 'rectangular')
        self.assertEqual(self.bounds_rectangular.lat_min, -80)
        self.assertEqual(self.bounds_rectangular.lat_max, 80)
        self.assertEqual(self.bounds_rectangular.lon_min, -160)
        self.assertEqual(self.bounds_rectangular.lon_max, 160)
        self.assertEqual(self.bounds_rectangular.start,
                         dt.datetime(2000, 1, 1))
        self.assertEqual(self.bounds_rectangular.end, dt.datetime(2002, 1, 1))

        self.assertEqual(self.bounds_CORDEX.boundary_type, 'CORDEX South Asia')
        self.assertEqual(self.bounds_CORDEX.lat_min, -15.23)
        self.assertEqual(self.bounds_CORDEX.lat_max, 45.07)
        self.assertEqual(self.bounds_CORDEX.lon_min, 19.88)
        self.assertEqual(self.bounds_CORDEX.lon_max, 115.55)

        self.assertEqual(self.bounds_us_states.boundary_type, 'us_states')

        self.assertEqual(self.bounds_countries.boundary_type, 'countries')

    # Temporal tests
    def test_inverted_start_end_times(self):
        with self.assertRaises(ValueError):
            self.bounds_rectangular.start = dt.datetime(2003, 1, 1)

        with self.assertRaises(ValueError):
            self.bounds_rectangular.end = dt.datetime(1999, 1, 1)

    # Start tests
    def test_invalid_start(self):
        with self.assertRaises(ValueError):
            self.bounds_rectangular.start = "This is not a date time object"

    # End tests
    def test_invalid_end(self):
        with self.assertRaises(ValueError):
            self.bounds_rectangular.end = "This is not a date time object"

    # Start tests
    def test_none_value_start(self):
        self.assertEqual(self.bounds_CORDEX.start, None)

    # End tests
    def test_none_value_end(self):
        self.assertEqual(self.bounds_CORDEX.end, None)

if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
OCW module to download a file from ESGF.

"""

from __future__ import print_function

import sys
from os.path import expanduser, join

from ocw.esgf.constants import ESGF_CREDENTIALS

if sys.version_info[0] >= 3:
    from http.client import HTTPSConnection
    from urllib.request import build_opener
    from urllib.request import HTTPCookieProcessor
    from urllib.request import HTTPSHandler
else:
    # Not Python 3 - today, it is most likely to be Python 2
    # But note that this might need an update when Python 4
    # might be around one day
    from httplib import HTTPSConnection
    from urllib2 import build_opener
    from urllib2 import HTTPCookieProcessor
    from urllib2 import HTTPSHandler


class HTTPSClientAuthHandler(HTTPSHandler):
    """
    HTTP handler that transmits an X509 certificate as part of the request
    """

    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        """
        Opens the https connection.
        :param req:  The https request object.
        :return: An addinfourl object for the request.
        """
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=300):
        """
        Create an HTTPSConnection object.
        :param host: The ESGF server to connect to.
        :param timeout: Connection timeout in seconds.
        :return:
        """
        return HTTPSConnection(host, key_file=self.key, cert_file=self.cert, timeout=timeout)


def download(url, toDirectory="/tmp"):
    """
    Function to download a single file from ESGF.
    :param url: the URL of the file to download
    :param toDirectory: target directory where the file will be written
    """

    # setup HTTP handler
    cert_file = expanduser(ESGF_CREDENTIALS)
    opener = build_opener(HTTPSClientAuthHandler(cert_file, cert_file))
    opener.add_handler(HTTPCookieProcessor())

    # download file
    local_file_path = join(toDirectory, url.split('/')[-1])
    print("\nDownloading url: %s to local path: %s ..." % (url, local_file_path))
    local_file = open(local_file_path, 'w')
    web_file = opener.open(url)
    local_file.write(web_file.read())

    # cleanup
    local_file.close()
    web_file.close()
    opener.close()
    print("... done")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''Unit test for the metrics.py module.'''

import unittest
import datetime as dt

from ocw.dataset import Dataset
import ocw.metrics as metrics

import numpy as np
import numpy.ma as ma
import numpy.testing as npt


class TestBias(unittest.TestCase):
    '''Test the metrics.Bias metric.'''

    def setUp(self):
        self.bias = metrics.Bias()
        # Initialize reference dataset
        self.reference_lat = np.array([10, 12, 14, 16, 18])
        self.reference_lon = np.array([100, 102, 104, 106, 108])
        self.reference_time = np.array(
            [dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.reference_value = flat_array.reshape(12, 5, 5)
        self.reference_variable = 'prec'
        self.reference_dataset = Dataset(self.reference_lat,
                                         self.reference_lon,
                                         self.reference_time,
                                         self.reference_value,
                                         self.reference_variable)
        # Initialize target dataset
        self.target_lat = np.array([1, 2, 4, 6, 8])
        self.target_lon = np.array([10, 12, 14, 16, 18])
        self.target_time = np.array(
            [dt.datetime(2001, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300, 600))
        self.target_value = flat_array.reshape(12, 5, 5)
        self.target_variable = 'tasmax'
        self.target_dataset = Dataset(self.target_lat,
                                      self.target_lon,
                                      self.target_time,
                                      self.target_value,
                                      self.target_variable)

    def test_function_run(self):
        '''Test bias function between reference dataset and target dataset.'''
        expected_result = np.zeros((12, 5, 5), dtype=np.int)
        expected_result.fill(-300)
        np.testing.assert_array_equal(self.bias.run(
            self.target_dataset, self.reference_dataset), expected_result)


class TestSpatialPatternTaylorDiagram(unittest.TestCase):
    '''Test the metrics.SpatialPatternTaylorDiagram'''

    def setUp(self):
        self.taylor_diagram = metrics.SpatialPatternTaylorDiagram()
        self.ref_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 5
            np.arange(0, 1500, 5).reshape(12, 5, 5),
            'ds1'
        )

        self.tar_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 2
            np.arange(0, 600, 2).reshape(12, 5, 5),
            'ds2'
        )

    def test_function_run(self):
        np.testing.assert_array_equal(self.taylor_diagram.run(
            self.ref_dataset, self.tar_dataset), ma.array([0.4, 1.0]))


class TestTemporalStdDev(unittest.TestCase):
    '''Test the metrics.TemporalStdDev metric.'''

    def setUp(self):
        self.temporal_std_dev = metrics.TemporalStdDev()
        # Initialize target dataset
        self.target_lat = np.array([10, 12, 14, 16, 18])
        self.target_lon = np.array([100, 102, 104, 106, 108])
        self.target_time = np.array(
            [dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.target_value = flat_array.reshape(12, 5, 5)
        self.target_variable = 'prec'
        self.target_dataset = Dataset(self.target_lat,
                                      self.target_lon,
                                      self.target_time,
                                      self.target_value,
                                      self.target_variable)

    def test_function_run(self):
        '''Test TemporalStdDev function for target dataset.'''
        expected_result = np.zeros((5, 5),)
        expected_result.fill(90.13878189)
        npt.assert_almost_equal(self.temporal_std_dev.run(
            self.target_dataset), expected_result)


class TestStdDevRatio(unittest.TestCase):
    '''Test the metrics.StdDevRatio metric'''

    def setUp(self):
        self.std_dev_ratio = metrics.StdDevRatio()
        self.ref_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 5
            np.arange(0, 1500, 5).reshape(12, 5, 5),
            'ds1'
        )

        self.tar_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 2
            np.arange(0, 600, 2).reshape(12, 5, 5),
            'ds2'
        )

    def test_function_run(self):
        self.assertTrue(self.std_dev_ratio.run(
            self.ref_dataset, self.tar_dataset), 0.4)


class TestPatternCorrelation(unittest.TestCase):
    '''Test the metrics.PatternCorrelation metric'''

    def setUp(self):
        self.pattern_correlation = metrics.PatternCorrelation()
        self.ref_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 5
            np.arange(0, 1500, 5).reshape(12, 5, 5),
            'ds1'
        )

        self.tar_dataset = Dataset(
            np.array([1., 1., 1., 1., 1.]),
            np.array([1., 1., 1., 1., 1.]),
            np.array([dt.datetime(2000, x, 1) for x in range(1, 13)]),
            # Reshapped array with 300 values incremented by 2
            np.arange(0, 600, 2).reshape(12, 5, 5),
            'ds2'
        )

    def test_function_run(self):
        pattern = self.pattern_correlation.run(
            self.tar_dataset, self.ref_dataset)
        self.assertEqual(pattern, 1.0)


class TestTemporalCorrelation(unittest.TestCase):
    '''Test the metrics.TemporalCorrelation metric.'''

    def setUp(self):
        # Set metric.
        self.metric = metrics.TemporalCorrelation()
        # Initialize reference dataset.
        self.ref_lats = np.array([10, 20, 30, 40, 50])
        self.ref_lons = np.array([5, 15, 25, 35, 45])
        self.ref_times = np.array([dt.datetime(2000, x, 1)
                                   for x in range(1, 13)])
        self.ref_values = np.array(range(300)).reshape(12, 5, 5)
        self.ref_variable = "ref"
        self.ref_dataset = Dataset(self.ref_lats,
                                   self.ref_lons,
                                   self.ref_times,
                                   self.ref_values,
                                   self.ref_variable)
        # Initialize target datasets.
        self.tgt_lats = np.array([10, 20, 30, 40, 50])
        self.tgt_lons = np.array([5, 15, 25, 35, 45])
        self.tgt_times = np.array([dt.datetime(2000, x, 1)
                                   for x in range(1, 13)])
        self.tgt_variable = "tgt"
        self.tgt_values_inc = np.array(range(300, 600)).reshape(12, 5, 5)
        self.tgt_values_dec = np.array(range(299, -1, -1)).reshape(12, 5, 5)
        self.tgt_dataset_inc = Dataset(self.tgt_lats,
                                       self.tgt_lons,
                                       self.tgt_times,
                                       self.tgt_values_inc,
                                       self.tgt_variable)
        self.tgt_dataset_dec = Dataset(self.tgt_lats,
                                       self.tgt_lons,
                                       self.tgt_times,
                                       self.tgt_values_dec,
                                       self.tgt_variable)

    def test_identical_inputs(self):
        expected = np.ones(25).reshape(5, 5)
        tc = self.metric.run(self.ref_dataset, self.ref_dataset)
        np.testing.assert_array_equal(tc, expected)

    def test_positive_correlation(self):
        expected = np.ones(25).reshape(5, 5)
        tc = self.metric.run(self.ref_dataset, self.tgt_dataset_inc)
        np.testing.assert_array_equal(tc, expected)

    def test_negative_correlation(self):
        expected_tc = np.array([-1] * 25).reshape(5, 5)
        tc = self.metric.run(self.ref_dataset, self.tgt_dataset_dec)
        np.testing.assert_array_equal(tc, expected_tc)


class TestTemporalMeanBias(unittest.TestCase):
    '''Test the metrics.TemporalMeanBias metric.'''

    def setUp(self):
        self.mean_bias = metrics.TemporalMeanBias()
        # Initialize reference dataset
        self.reference_lat = np.array([10, 12, 14, 16, 18])
        self.reference_lon = np.array([100, 102, 104, 106, 108])
        self.reference_time = np.array(
            [dt.datetime(2000, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300))
        self.reference_value = flat_array.reshape(12, 5, 5)
        self.reference_variable = 'prec'
        self.reference_dataset = Dataset(self.reference_lat,
                                         self.reference_lon,
                                         self.reference_time,
                                         self.reference_value,
                                         self.reference_variable)
        # Initialize target dataset
        self.target_lat = np.array([1, 2, 4, 6, 8])
        self.target_lon = np.array([10, 12, 14, 16, 18])
        self.target_time = np.array(
            [dt.datetime(2001, x, 1) for x in range(1, 13)])
        flat_array = np.array(range(300, 600))
        self.target_value = flat_array.reshape(12, 5, 5)
        self.target_variable = 'tasmax'
        self.target_dataset = Dataset(self.target_lat,
                                      self.target_lon,
                                      self.target_time,
                                      self.target_value,
                                      self.target_variable)

    def test_function_run(self):
        '''
        Test mean bias function between reference dataset and target dataset.
        '''
        expected_result = np.zeros((5, 5), dtype=np.int)
        expected_result.fill(-300)
        np.testing.assert_array_equal(self.mean_bias.run(
            self.target_dataset, self.reference_dataset), expected_result)


class TestRMSError(unittest.TestCase):
    '''Test the metrics.RMSError metric.'''

    def setUp(self):
        # Set metric.
        self.metric = metrics.RMSError()
        # Initialize reference dataset.
        self.ref_lats = np.array([10, 20, 30, 40, 50])
        self.ref_lons = np.array([5, 15, 25, 35, 45])
        self.ref_times = np.array([dt.datetime(2000, x, 1)
                                   for x in range(1, 13)])
        self.ref_values = np.array([4] * 300).reshape(12, 5, 5)
        self.ref_variable = "ref"
        self.ref_dataset = Dataset(self.ref_lats,
                                   self.ref_lons,
                                   self.ref_times,
                                   self.ref_values,
                                   self.ref_variable)
        # Initialize target dataset.
        self.tgt_lats = np.array([10, 20, 30, 40, 50])
        self.tgt_lons = np.array([5, 15, 25, 35, 45])
        self.tgt_times = np.array([dt.datetime(2000, x, 1)
                                   for x in range(1, 13)])
        self.tgt_values = np.array([2] * 300).reshape(12, 5, 5)
        self.tgt_variable = "tgt"
        self.tgt_dataset = Dataset(self.tgt_lats,
                                   self.tgt_lons,
                                   self.tgt_times,
                                   self.tgt_values,
                                   self.tgt_variable)

    def test_function_run(self):
        result = self.metric.run(self.ref_dataset, self.tgt_dataset)
        self.assertEqual(result, 2.0)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
import os
import netCDF4
import numpy as np
from ocw.dataset import Dataset
from ocw.dataset_loader import DatasetLoader


class TestDatasetLoader(unittest.TestCase):

    def setUp(self):
        # Read netCDF file
        self.file_path = create_netcdf_object()
        self.netCDF_file = netCDF4.Dataset(self.file_path, 'r')
        self.latitudes = self.netCDF_file.variables['latitude'][:]
        self.longitudes = self.netCDF_file.variables['longitude'][:]
        self.times = self.netCDF_file.variables['time'][:]
        self.alt_lats = self.netCDF_file.variables['alt_lat'][:]
        self.alt_lons = self.netCDF_file.variables['alt_lon'][:]
        self.values = self.netCDF_file.variables['value'][:]
        self.values2 = self.values + 1

        # Set up config
        self.config = {'file_path': self.file_path, 'variable_name': 'value'}
        self.new_data_source_config = {'loader_name': 'foo',
                                       'lats': self.latitudes,
                                       'lons': self.longitudes,
                                       'times': self.times,
                                       'values': self.values2,
                                       'variable': 'value'}

    def tearDown(self):
        os.remove(self.file_path)

    def testNewDataSource(self):
        '''
        Ensures that custom data source loaders can be added
        '''
        self.loader = DatasetLoader(self.new_data_source_config)

        # Here the data_source "foo" represents the Dataset constructor
        self.loader.add_source_loader('foo', build_dataset)
        self.loader.load_datasets()
        self.assertEqual(self.loader.datasets[0].origin['source'], 'foo')
        np.testing.assert_array_equal(self.loader.datasets[0].values,
                                      self.values2)

    def testExistingDataSource(self):
        '''
        Ensures that existing data source loaders can be added
        '''
        self.loader = DatasetLoader(self.config)
        self.loader.load_datasets()
        self.assertEqual(self.loader.datasets[0].origin['source'], 'local')
        np.testing.assert_array_equal(self.loader.datasets[0].values,
                                      self.values)

    def testMultipleDataSources(self):
        '''
        Test for when multiple dataset configs are specified
        '''
        self.loader = DatasetLoader(self.config, self.new_data_source_config)

        # Here the data_source "foo" represents the Dataset constructor
        self.loader.add_source_loader('foo', build_dataset)
        self.loader.load_datasets()
        self.assertEqual(self.loader.datasets[0].origin['source'],
                         'local')
        self.assertEqual(self.loader.datasets[1].origin['source'],
                         'foo')
        np.testing.assert_array_equal(self.loader.datasets[0].values,
                                      self.values)
        np.testing.assert_array_equal(self.loader.datasets[1].values,
                                      self.values2)


def build_dataset(*args, **kwargs):
    '''
    Wrapper to Dataset constructor from fictitious 'foo' data_source.
    '''
    origin = {'source': 'foo'}
    return Dataset(*args, origin=origin, **kwargs)


def create_netcdf_object():
    # To create the temporary netCDF file
    file_path = '/tmp/temporaryNetcdf.nc'
    netCDF_file = netCDF4.Dataset(file_path, 'w', format='NETCDF4')
    # To create dimensions
    netCDF_file.createDimension('lat_dim', 5)
    netCDF_file.createDimension('lon_dim', 5)
    netCDF_file.createDimension('time_dim', 3)
    # To create variables
    latitudes = netCDF_file.createVariable('latitude', 'd', ('lat_dim',))
    longitudes = netCDF_file.createVariable('longitude', 'd', ('lon_dim',))
    times = netCDF_file.createVariable('time', 'd', ('time_dim',))
    # unusual variable names to test optional arguments for Dataset constructor
    alt_lats = netCDF_file.createVariable('alt_lat', 'd', ('lat_dim',))
    alt_lons = netCDF_file.createVariable('alt_lon', 'd', ('lon_dim',))
    alt_times = netCDF_file.createVariable('alt_time', 'd', ('time_dim',))
    values = netCDF_file.createVariable('value', 'd',
                                        ('time_dim',
                                         'lat_dim',
                                         'lon_dim')
                                        )

    # To latitudes and longitudes for five values
    latitudes_data = np.arange(5.)
    longitudes_data = np.arange(150., 155.)
    # Three months of data.
    times_data = np.arange(3)
    # Create 150 values
    values_data = np.array([i for i in range(75)])
    # Reshape values to 4D array (level, time, lats, lons)
    values_data = values_data.reshape(len(times_data), len(latitudes_data),
                                      len(longitudes_data))

    # Ingest values to netCDF file
    latitudes[:] = latitudes_data
    longitudes[:] = longitudes_data
    times[:] = times_data
    alt_lats[:] = latitudes_data + 10
    alt_lons[:] = longitudes_data - 10
    alt_times[:] = times_data
    values[:] = values_data
    # Assign time info to time variable
    netCDF_file.variables['time'].units = 'months since 2001-01-01 00:00:00'
    netCDF_file.variables[
        'alt_time'].units = 'months since 2001-04-01 00:00:00'
    netCDF_file.variables['value'].units = 'foo_units'
    netCDF_file.close()
    return file_path

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

'''
Classes:
    RCMED - A class for retrieving data from Regional Climate Model Evalutaion Database (JPL).
    More information about the RCMED Query Specification can be found below:
    https://rcmes.jpl.nasa.gov/query-api/query.php?
'''
# Needed Python 2/3 urllib compatability
try:
    from urllib.parse import urlencode
    from urllib.request import urlopen
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen

import re
import json
import numpy as np
import numpy.ma as ma
from datetime import datetime
import calendar
from ocw.dataset import Dataset


URL = 'http://rcmes.jpl.nasa.gov/query-api/query.php?'


def get_parameters_metadata():
    '''Get the metadata of all parameter from RCMED.

    :returns: Dictionary of information for each parameter stored in one list
    :rtype: :class:`list` of :class:`dict`
    '''

    param_info_list = []
    url = URL + "&param_info=yes"
    string = urlopen(url)
    data_string = string.read().decode('utf-8')
    json_format_data = json.loads(data_string)
    fields_name = json_format_data['fields_name']
    data = json_format_data['data']
    for row in data:
        dic = {}
        for name in fields_name:
            dic[name] = row[fields_name.index(name)]
        param_info_list.append(dic)

    return param_info_list


def _make_mask_array(values, parameter_id, parameters_metadata):
    '''Created masked array to deal with missing values

    :param values: Numpy array of values which may contain missing values
    :type values: Numpy array
    :param parameter_id: Parameter's id
    :type parameter_id: Integer
    :param parameters_metadata: Metadata for all parameters
    :type parameters_metadata: List of dictionaries

    :returns: Masked array of values
    :rtype: Masked array
    '''

    for each in parameters_metadata:
        if str(each['parameter_id']) == str(parameter_id):
            missing_values = each['missingdataflag'].encode()
            break
    missing_values = float(missing_values)
    values = ma.masked_array(values, mask=(values == missing_values))

    return values


def _reshape_values(values, unique_values):
    '''Reshape values into 4D array.

    :param values: Raw values data
    :type values: numpy array
    :param unique_values: Tuple of unique latitudes, longitudes and times data.
    :type unique_values: Tuple

    :returns: Reshaped values data
    :rtype: Numpy array
    '''

    lats_len = len(unique_values[0])
    lons_len = len(unique_values[1])
    times_len = len(unique_values[2])

    values = values.reshape(times_len, lats_len, lons_len)

    return values


def _calculate_time(unique_times, time_step):
    '''Convert each time to the datetime object.

    :param unique_times: Unique time data
    :type unique_times: String
    :param time_step: Time step
    :type time_step: String

    :returns: Unique datetime objects of time data
    :rtype: Numpy array
    '''

    time_format = "%Y-%m-%d %H:%M:%S"
    unique_times = np.array(
        [datetime.strptime(time.decode('utf-8'), time_format)
                           for time in unique_times])
    # There is no need to sort time.
    # This function may required still in RCMES
    # unique_times.sort()
    # This function should be moved to the data_process.

    return unique_times


def _make_unique(lats, lons, times):
    '''Find the unique values of input data.

    :param lats: lats
    :type lats: Numpy array
    :param lons: lons
    :type lons: Numpy array
    :param times: times
    :type times: Numpy array

    :returns: Unique numpy arrays of latitudes, longitudes and times
    :rtype: Tuple
    '''

    unique_lats = np.unique(lats)
    unique_lons = np.unique(lons)
    unique_times = np.unique(times)

    return (unique_lats, unique_lons, unique_times)


def _get_data(url):
    '''Reterive data from database.

    :param url: url to query from database
    :type url: String

    :returns: Latitudes, longitudes, times and values data
    :rtype: (Numpy array, Numpy array, Numpy array, Numpy array)
    '''

    string = urlopen(url)
    data_string = string.read()
    index_of_data = re.search(b'data: \r\n', data_string)
    data = data_string[index_of_data.end():len(data_string)]
    data = data.split(b'\r\n')

    lats = []
    lons = []
    #levels = []
    values = []
    times = []

    # Because the last row is empty, "len(data)-1" is used.
    for i in range(len(data) - 1):
        row = data[i].split(b',')
        lats.append(np.float32(row[0]))
        lons.append(np.float32(row[1]))
        # Level is not currently supported in Dataset class.
        # levels.append(np.float32(row[2]))
        times.append(row[3])
        values.append(np.float32(row[4]))

    lats = np.array(lats)
    lons = np.array(lons)
    times = np.array(times)
    values = np.array(values)

    return lats, lons, times, values


def _beginning_of_date(time, time_step):
    '''Calculate the beginning of given time, based on time step.

    :param time: Given time
    :type time: Datetime
    :param time_step: Time step (monthly or daily)
    :type time_step: String

    :returns: Beginning of given time
    :rtype: Datetime
    '''

    if time_step.lower() == 'monthly':
        if time.day != 1:
            time = datetime(time.year, time.month, 1)
    elif time_step.lower() == 'daily':
        if time.hour != 0 or time.minute != 0 or time.second != 0:
            time = datetime(time.year, time.month, time.day, 00, 00, 00)

    return time


def _end_of_date(time, time_step):
    '''Calculate the end of given time, based on time step.

    :param time: Given time
    :type time: Datetime
    :param time_step: Time step (monthly or daily)
    :type time_step: String

    :returns: End of given time
    :rtype: Datetime
    '''

    last_day_of_month = calendar.monthrange(time.year, time.month)[1]
    if time_step.lower() == 'monthly':
        time = datetime(time.year, time.month, last_day_of_month)
    elif time_step.lower() == 'daily':
        time = datetime(time.year, time.month, time.day, 23, 59, 59)

    return time


def _generate_query_url(dataset_id, parameter_id, min_lat, max_lat, min_lon, max_lon, start_time, end_time, time_step):
    '''Generate the url to query from database

    :param dataset_id: Dataset id.
    :type dataset_id: Integer
    :param parameter_id: Parameter id
    :type parameter_id: Integer
    :param min_lat: Minimum latitude
    :type min_lat: Float
    :param max_lat: Maximum latitude
    :type max_lat: Float
    :param min_lon: Minimum longitude
    :type min_lon: Float
    :param max_lon: Maximum longitude
    :type max_lon: Float
    :param start_time: Start time
    :type start_time: Datetime
    :param end_time: End time
    :type end_time: Datetime
    :param time_step: Time step
    :type time_step: String

    :returns: url to query from database
    :rtype: String
    '''

    start_time = _beginning_of_date(start_time, time_step)
    end_time = _end_of_date(end_time, time_step)
    start_time = start_time.strftime("%Y%m%dT%H%MZ")
    end_time = end_time.strftime("%Y%m%dT%H%MZ")

    query = [('datasetId', dataset_id), ('parameterId', parameter_id), ('latMin', min_lat), ('latMax', max_lat),
             ('lonMin', min_lon), ('lonMax', max_lon), ('timeStart', start_time), ('timeEnd', end_time)]

    query_url = urlencode(query)
    url_request = URL + query_url

    return url_request


def _get_parameter_info(parameters_metadata, parameter_id):
    '''General information for given parameter id.

    :param parameters_metadata: Metadata for all parameters
    :type parameters_metadata: List of dictionaries
    :param parameter_id: Parameter id
    :type parameter_id: Integer

    :returns: Database name, time step, realm, instrument, start_date, end_date and unit for given parameter
    :rtype: (string, string, string, string, string, string, string)
    '''

    for dic in parameters_metadata:
        if int(dic['parameter_id']) == parameter_id:
            database = dic['database']
            time_step = dic['timestep']
            realm = dic['realm']
            instrument = dic['instrument']
            start_date = dic['start_date']
            end_date = dic['end_date']
            unit = dic['units']

    return (database, time_step, realm, instrument, start_date, end_date, unit)


def parameter_dataset(dataset_id, parameter_id, min_lat, max_lat, min_lon, max_lon, start_time, end_time, name=''):
    '''Get data from one database(parameter).

    :param dataset_id: Dataset id.
    :type dataset_id: :class:`int`

    :param parameter_id: Parameter id
    :type parameter_id: :class:`int`

    :param min_lat: Minimum latitude
    :type min_lat: :class:`float`

    :param max_lat: Maximum latitude
    :type max_lat: :class:`float`

    :param min_lon: Minimum longitude
    :type min_lon: :class:`float`

    :param max_lon: Maximum longitude
    :type max_lon: :class:`float`

    :param start_time: Start time
    :type start_time: :class:`datetime.datetime`

    :param end_time: End time
    :type end_time: :class:`datetime.datetime`

    :param name: (Optional) A name for the loaded dataset.
    :type name: :mod:`string`

    :returns: An OCW Dataset object contained the requested data from RCMED.
    :rtype: :class:`dataset.Dataset`
    '''

    parameters_metadata = get_parameters_metadata()
    parameter_name, time_step, _, _, _, _, parameter_units = _get_parameter_info(
        parameters_metadata, parameter_id)

    lats, lons, times, values = \
        _coalesce_data(dataset_id, parameter_id, min_lat, max_lat, min_lon, max_lon,
                       start_time, end_time, time_step)

    unique_lats_lons_times = _make_unique(lats, lons, times)
    unique_times = _calculate_time(unique_lats_lons_times[2], time_step)
    values = _reshape_values(values, unique_lats_lons_times)
    values = _make_mask_array(values, parameter_id, parameters_metadata)

    origin = {
        'source': 'rcmed',
        'dataset_id': dataset_id,
        'parameter_id': parameter_id
    }

    return Dataset(unique_lats_lons_times[0],
                   unique_lats_lons_times[1],
                   unique_times,
                   values,
                   variable=parameter_name,
                   units=parameter_units,
                   name=name,
                   origin=origin)


def _coalesce_data(dataset_id, parameter_id, min_lat, max_lat, min_lon, max_lon,
                       start_time, end_time, time_step):

    """
    Refer to this JIRA:  https://issues.apache.org/jira/browse/CLIMATE-744

    Sometimes RCMED does not seem to return the entire data set when the requested
    range of data and / or number of data points are very large.  This method breaks
    the single large query into several smaller queries and then appends the results.

    :param dataset_id:  The RCMED dataset ID.
    :param parameter_id:  The parameter ID within the RCMED dataset.
    :param min_lat: The minimum lat of the dataset boundary.
    :param max_lat: The maximum lat of the dataset boundary.
    :param min_lon: The minimum lon of the dataset boundary.
    :param max_lon: The maximum lon of the dataset boundary.
    :param start_time: The start datetime of the dataset boundary.
    :param end_time: The end datetime of the dataset boundary.
    :param time_step: The timestep to use when segmenting the datetime boundary.
    :return:  lats, lons, times, and values for the requested dataset / parameter from RCMED.
    """

    lats = None
    lons = None
    times = None
    values = None

    # This is a magic number which strikes a balance between making an excessive number of
    # calls to RCMED (e.g. 1) and RCMED not sending back the full data set.
    step = 4

    current_start = start_time
    current_end = min(end_time, datetime(current_start.year + step, 12, 31))

    while True:

        url = _generate_query_url(dataset_id, parameter_id, min_lat,
                                  max_lat, min_lon, max_lon, current_start, current_end, time_step)

        tmp_lats, tmp_lons, tmp_times, tmp_values = _get_data(url)

        if lats is None:
            lats = tmp_lats
        else:
            lats = np.append(lats, tmp_lats)

        if lons is None:
            lons = tmp_lons
        else:
            lons = np.append(lons, tmp_lons)

        if times is None:
            times = tmp_times
        else:
            times = np.append(times, tmp_times)

        if values is None:
            values = tmp_values
        else:
            values = np.append(values, tmp_values)

        if current_end == end_time:
            break

        current_start = datetime(current_end.year + 1, 1, 1)
        current_end = min(end_time, datetime(current_start.year + step, 12, 31))


    return lats, lons, times, values
